
__author__    = "Andre Merzky, Mark Santcroos, Ole Weidner"
__copyright__ = "Copyright 2012-2015, The SAGA Project"
__license__   = "MIT"


""" Condor job adaptor implementation
"""

import re
import os
import time
import datetime

from urllib.parse import parse_qs
from tempfile import NamedTemporaryFile

import radical.utils as ru

from .. import base
from .. import cpi

from ...exceptions    import *
from ...job.constants import *
from ...exceptions    import *
from ...              import job        as api
from ...              import filesystem as sfs
from ...utils         import pty_shell  as sups
from ...utils.job     import TransferDirectives

SYNC_CALL  = cpi.decorators.SYNC_CALL
ASYNC_CALL = cpi.decorators.ASYNC_CALL


# --------------------------------------------------------------------
#
def _condor_to_saga_jobstate(condorjs):
    """ translates a condor one-letter state to saga
    """
    # From: http://pages.cs.wisc.edu/~adesmet/status.html
    #
    # JobStatus in job ClassAds
    #
    # 0   Unexpanded      U
    # 1   Idle            I
    # 2   Running         R
    # 3   Removed         X
    # 4   Completed       C
    # 5   Held            H
    # 6   Submission_err  E

    if   int(condorjs) == 0: return api.PENDING
    elif int(condorjs) == 1: return api.PENDING
    elif int(condorjs) == 2: return api.RUNNING
    elif int(condorjs) == 3: return api.CANCELED
    elif int(condorjs) == 4: return api.DONE
    elif int(condorjs) == 5: return api.PENDING
    elif int(condorjs) == 6: return api.FAILED
    else                   : return api.UNKNOWN


# --------------------------------------------------------------------
#
def _condorscript_generator(url, logger, jds, option_dict=None):
    """
    generates a Condor script from a set of SAGA job descriptions
    """

    if not isinstance(jds, list):
        jds = [jds]

    # assert that all jds belong to the same project (which might be empty)
    project = jds[0].project
    for jd in jds:
        assert(project == jd.project)


    condor_file = str()

    # HTCondor quoting/escaping:
    # http://research.cs.wisc.edu/htcondor/manual/current/condor_submit.html#SECTION0012514000000000000000

    ##### options passed via job service url #####
    if project or option_dict:
        condor_file += "\n##### OPTIONS PASSED VIA JOB SERVICE URL #####\n"

        if project:
            condor_file += "\n+ProjectName = \"%s\"" % jd.project

        if option_dict:
            for key,value in option_dict.items():
                condor_file += "\n%s = %s" % (key, value)


    ##### options passed via job description #####
    condor_file += "\n\n##### OPTIONS PASSED VIA JOB DESCRIPTION #####\n"

    # Per Job in Bulk settings
    for jd in jds:

        # environment -> environment
        # http://research.cs.wisc.edu/htcondor/manual/current/condor_submit.html#SECTION0012514000000000000000
        if jd.environment:
            environment = ''
            for key,val in jd.environment.items():
                environment += ' "%s=%s"'     % (key, val)
              # condor_file += "\n%-25s = %s" % (key, val)
            condor_file += "\nenvironment               = %s\n" % environment

        condor_file += "\nstream_output             = True"
        condor_file += "\nstream_error              = True"
        condor_file += "\nshould_transfer_files     = YES"
        condor_file += "\nwhen_to_transfer_output   = ON_EXIT_OR_EVICT\n"


        # handle site inclusion/exclusion
        requirements = "(NumJobStarts =?= 0 || NumJobStarts =?= Undefined)"
      # requirements = "(Target.MAX.PREEMPT >= 24*60*60) || (Target.MAX.PREEMPT == false)"

        if jd.candidate_hosts:

            # special characters / strings
            # TODO: this is bound to GLIDEIN which it shouldnt be.
            EXCLUSION    = '!'
            REQUIRED     = '~'
            KEY          = 'GLIDEIN_ResourceName'

            # Whitelist sites, filter out "special" entries from the candidate
            # host lists
            excl_sites = [host for host in jd.candidate_hosts
                               if  host.startswith(EXCLUSION)]
            req_sites  = [host for host in jd.candidate_hosts
                               if  host.startswith(REQUIRED)]
            incl_sites = [host for host in jd.candidate_hosts
                               if not host.startswith((REQUIRED, EXCLUSION))]

            if incl_sites:
                requirements += ' && (' +  ' || '.join(['%s =?= "%s"' % (KEY, site)
                                for site in incl_sites]) + ')'

            # Blacklist sites, strip out leading '!'
            if excl_sites:
                requirements += ' && (' +  ' && '.join(['%s =!= "%s"' % (KEY, site[1:])
                        for site in excl_sites]) + ')'

            # Get the '~special_requirements' and strip leading ~
            if req_sites:
                requirements += ' && '.join(req_sites)

        condor_file += "\nrequirements    = %s" % requirements
        condor_file += "\nperiodic_remove = NumJobStarts > 0 && JobStatus == 1\n"


        # special treatment for universe - defaults to 'vanilla'
        if jd.queue:
            condor_file += "\nuniverse     = %s" % jd.queue
        else:
            condor_file += "\nuniverse     = vanilla"

        # Condor doesn't expand environment variables in arguments.
        # To support this functionality, we wrap by calling
        #     /bin/env /bin/sh -c ""
        # where the arg string to the shell gets expanded
        condor_file += "\nexecutable   = /bin/env"

        # the actual executable becomes the first argument.
        if not jd.executable:
            raise NoSuccess("Executable not set")
        arguments = "%s" % jd.executable

        # all other arguments follow
        if jd.arguments:

            for arg in jd.arguments:

                # Condor can't deal with multi-line arguments. yep, that's how
                # bad of a software it is.
                if '\n' in arg:
                    raise NoSuccess("Condor doesn't support multi-line arguments: %s" % arg)

                # Condor HATES double quotes in the arguments. It'll return
                # some crap like: "Found illegal unescaped double-quote: ...
                # That's why we escape them by repeating.
                arg = arg.replace('"', '""')
                arg = arg.replace("'", "''")

                # Escape dollars (for environment variables)
                #arg = arg.replace('$', '\\$')

                arguments += " %s" % arg

        # arguments -> arguments
        condor_file += "\narguments    = \"/bin/sh -c '%s'\"" % arguments

        # Transfer Directives
        assert(jd.transfer_directives)

        # all checking is done in _handle_file_transfers() already.
        condor_file += "\n"
        td = jd.transfer_directives

        if td.transfer_output_files:
            condor_file += "\ntransfer_output_files = %s" \
                         % ','.join(td.transfer_output_files)

        if td.transfer_input_files:
            condor_file += "\ntransfer_input_files  = %s\n" \
                         % ','.join(td.transfer_input_files)

        # TODO: what to do when working directory is not set?
        job_pwd = './'
        if jd.working_directory:
            job_pwd = jd.working_directory
            condor_file += "\ninitialdir   = %s " % job_pwd
        logname = "radical-saga-job-$(cluster)_$(process).log"
        condor_file += "\nlog          = %s " % os.path.join(job_pwd, logname)

        # output -> output
        if jd.output is not None:
            condor_file += "\noutput       = %s " % os.path.join(job_pwd, jd.output)

        # error -> error
        if jd.error is not None:
            condor_file += "\nerror        = %s " % os.path.join(job_pwd, jd.error)

        if jd.total_cpu_count:
            condor_file += "\nrequest_cpus = %d" % jd.total_cpu_count


        # 'queue' concludes the description of a job
        condor_file += "\nqueue\n"
        condor_file += "\n##### END OF JOB #####\n\n"

    condor_file += "\n##### END OF FILE #####\n\n"

    return condor_file


# --------------------------------------------------------------------
# some private defs
#
_CACHE_TIMEOUT = 5.0

# --------------------------------------------------------------------
# the adaptor name
#
_ADAPTOR_NAME          = "radical.saga.adaptors.condorjob"
_ADAPTOR_SCHEMAS       = ["condor", "condor+ssh", "condor+gsissh"]

# --------------------------------------------------------------------
# the adaptor capabilities & supported attributes
#
_ADAPTOR_CAPABILITIES = {
    "jdes_attributes":   [NAME,
                          EXECUTABLE,
                          ARGUMENTS,
                          ENVIRONMENT,
                          INPUT,
                          OUTPUT,
                          ERROR,
                          QUEUE, # TODO: map jd.queue to universe or pool?!
                          PROJECT,
                          WALL_TIME_LIMIT,
                          WORKING_DIRECTORY,
                          CANDIDATE_HOSTS,
                          TOTAL_CPU_COUNT,
                          PROCESSES_PER_HOST,
                          SPMD_VARIATION,
                          FILE_TRANSFER],
    "job_attributes":    [EXIT_CODE,
                          EXECUTION_HOSTS,
                          NAME,
                          CREATED,
                          STARTED,
                          FINISHED],
    "metrics":           [STATE],
    "contexts":          {"ssh": "SSH public/private keypair",
                          "x509": "GSISSH X509 proxy context",
                          "userpass": "username/password pair (ssh)"}
}

# --------------------------------------------------------------------
# the adaptor documentation
#
_ADAPTOR_DOC = {
    "name":          _ADAPTOR_NAME,
    "capabilities":  _ADAPTOR_CAPABILITIES,
    "description":  """
The (HT)Condor(-G) adaptor allows to run and manage jobs on a
`Condor <http://research.cs.wisc.edu/htcondor/>`_ gateway.
""",
    "example": "examples/jobs/condorjob.py",
    "schemas": {"condor"        : "connect to a local gateway",
                "condor+ssh"    : "connect to a remote gateway via SSH",
                "condor+gsissh ": "connect to a remote gateway via GSISSH"}
}

# --------------------------------------------------------------------
# the adaptor info is used to register the adaptor with SAGA
#
_ADAPTOR_INFO = {
    "name"             : _ADAPTOR_NAME,
    "version"          : "v0.2",
    "schemas"          : _ADAPTOR_SCHEMAS,
    "capabilities"     : _ADAPTOR_CAPABILITIES,
    "cpis": [
        {
        "type"         : "radical.saga.job.Service",
        "class"        : "CondorJobService"
        },
        {
        "type"         : "radical.saga.job.Job",
        "class"        : "CondorJob"
        }
    ]
}


###############################################################################
# The adaptor class
class Adaptor (base.Base):
    """ this is the actual adaptor class, which gets loaded by SAGA (i.e. by
        the SAGA engine), and which registers the CPI implementation classes
        which provide the adaptor's functionality.
    """

    # ----------------------------------------------------------------
    #
    def __init__(self):

        base.Base.__init__(self, _ADAPTOR_INFO)

        self.id_re    = re.compile('^\[(.*)\]-\[(.*?)\]$')
        self.use_hist = self._cfg.get('use_history', False)
        self.epoch    = datetime.datetime(1970,1,1)

        self._logger.info("use condor_history: %s", self.use_hist)


    # ----------------------------------------------------------------
    #
    def sanity_check(self):
        # FIXME: also check for gsissh
        pass

    # ----------------------------------------------------------------
    #
    def parse_id(self, id):
        # split the id '[rm]-[pid]' in its parts, and return them.

        match = self.id_re.match(id)

        if not match or len(match.groups()) != 2:
            raise BadParameter("Cannot parse job id '%s'" % id)

        return (match.group(1), match.group(2))


###############################################################################
#
class CondorJobService (cpi.job.Service):
    """ implements cpi.job.Service
    """

    # ----------------------------------------------------------------
    #
    def __init__(self, api, adaptor):

        _cpi_base = super(CondorJobService, self)
        _cpi_base.__init__(api, adaptor)

        self._adaptor = adaptor


    # ----------------------------------------------------------------
    #
    def __del__(self):

        self.finalize(kill_shell=True)

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def init_instance(self, adaptor_state, rm_url, session):
        """ service instance constructor
        """
        self.rm            = rm_url
        self.session       = session
        self.ppn           = 0
        self.is_cray       = False
        self.jobs          = dict()
        self.query_options = dict()

        rm_scheme = rm_url.scheme
        pty_url   = ru.Url (rm_url)

        # this adaptor supports options that can be passed via the
        # 'query' component of the job service URL.
        if rm_url.query is not None:
            for key, val in parse_qs(rm_url.query).items():
                self.query_options[key] = val[0]

        # we need to extract the scheme for PTYShell. That's basically the
        # job.Service Url without the condor+ part. We use the PTYShell to
        # execute condor commands either locally or via gsissh or ssh.
        if rm_scheme == "condor":
            pty_url.scheme = "fork"
        elif rm_scheme == "condor+ssh":
            pty_url.scheme = "ssh"
        elif rm_scheme == "condor+gsissh":
            pty_url.scheme = "gsissh"

        # these are the commands that we need in order to interact with Condor.
        # the adaptor will try to find them during initialize(self) and bail
        # out in case they are not available.
        self._commands = {'condor_version': None,
                          'condor_submit':  None,
                          'condor_q':       None,
                          'condor_history': None,
                          'condor_vacate':  None,
                          'condor_rm':      None}

        self.shell = sups.PTYShell(pty_url, self.session)

      # self.shell.set_initialize_hook(self.initialize)
      # self.shell.set_finalize_hook(self.finalize)

        self.initialize()

        return self.get_api ()


    # ----------------------------------------------------------------
    #
    def close (self) :
        if  self.shell :
            self.shell.finalize (True)


    # ----------------------------------------------------------------
    #
    def initialize(self):

        # check if all required condor tools are available
        commands = list(self._commands.keys())
        ret, out, _ = self.shell.run_sync("which %s " % ' '.join(commands))

        if ret != 0:
            raise NoSuccess("Error finding Condor tools: %s" % out)

        # split and remove empty lines
        lines = list(filter(bool, out.split('\n')))

        if len(lines) != len(commands):
            raise NoSuccess("Error finding some Condor tools: %s" % out)

        for cmd, path in zip(commands, lines):
            self._commands[cmd] = path.strip()

          # if cmd == 'condor_version':
          #     ret, out, _ = self.shell.run_sync("%s" % cmd)
          #     if ret != 0:
          #         raise NoSuccess("Error getting Condor version: %s" % out)
          #     else:
          #         # version is reported as:
          #         # $CondorVersion: 7.8.6 Oct 25 2012 $
          #         # $CondorPlatform: X86_64-CentOS_5.7 $
          #         lines = out.split('\n')
          #         version = lines[0].replace("$CondorVersion: ", "")
          #         version = version.strip(" $")
          #
          #         # add path and version to the command dictionary
          # self._commands[cmd] = {"path":    path}
          #                        "version": version}

        self._logger.info("Found Condor tools: %s", self._commands)

    # ----------------------------------------------------------------
    #
    def finalize(self, kill_shell=False):
        pass


    # --------------------------------------------------------------------------
    #
    # condor_q can fail due to the backend timing out, or other backend errors
    # in general.  We thus use this wrapper around condor_q to retry a couple of
    # times in those events.
    #
    # TODO: make the decision to retry somewhat more clever, like only on
    #       timeouts etc.
    #
    def _run_condor_q(self, retries=1, timeout=10,
                      pid=None, options=None, egrep=None):
        ret = None
        out = ""
        err = ""
        cmd = 'unset GREP_OPTIONS; %s' % self._commands['condor_q']

        if pid:     cmd += ' -long %s'       % pid
        if options: cmd += ' %s'             % options

        if egrep:
            cmd += " | grep"
            for expr in egrep:
                cmd += " -e '%s'" % expr

        # we  try at most 'retries' times - but the loop may exit early
        for n in range(retries):

            # if we tried before wait for a little
            if ret != None and timeout:
                time.sleep(timeout)

            _ret, _out, _err = self.shell.run_sync(cmd)
            if _ret == 0:
                # this one worked - ignore any previous runs (if any)
                # and record the result
                ret = _ret
                out = _out
                err = _err
                break

            else:
                # this one failed - record error and retval
                ret  = _ret
                out += '\n\nretry %d:\n%s' % (n, _out)
                err += '\n\nretry %d:\n%s' % (n, _err)

                self._logger.debug('need retry %s', options)

                # if pid was given, then failur means job not found.  No point
                # retrying...
                if pid:
                    break

        # we exhausted the number of retries, or we succeeded - either way,
        # return what we have.  This will return the most recent retval, and
        # output.  In the case of repeated errors, the output contains
        # a *concatenation* of error messages from the different tries.
        return ret, out, err


    # ----------------------------------------------------------------
    #
    def _new_job_info(self):
        return {'state':        UNKNOWN,
                'exec_hosts':   None,
                'returncode':   None,
                'create_time':  None,
                'start_time':   None,
                'end_time':     None,
                'gone':         False,
                'td':           TransferDirectives(),
                'stdout':       None,
                'stderr':       None,
                'name':         None,
                'timestamp':    0.0
            }

    # --------------------------------------------------------------------------
    #
    def _prepare_jd(self, jd):

        if not jd.file_transfer:
            jd.file_transfer = []

        # Our submitted Condor script wraps the executable into a /bin/sh
        # command, so Condor will not be able to infer the actual executable,
        # and thus will make no attempt at staging it to the target host --
        # which is what Condor usually does.
        #
        # To recover that behavior, we add the respective file transfer
        # directive for the second hop (first hop: laptop -> submission host,
        # second hop : submission host -> execution site).  However, we only add
        # that for executables starting with './', as we can't stage to other
        # locations than the job's workdir anyway.
        #
        # If a SAGA application wants to stage the executable over the *first*
        # hop, ie. to the submission host, it has to explicitly specify that in
        # the job description -- just like SAGA requires for all other backends.
        if jd.executable.startswith('./'):

            exe = jd.executable

            # Check if the executable is already in the file_transfer list,
            # because then we don't need to implicitly add it anymore.
            # (For example, if the executable is not in the local directory, it needs
            # to be explicitly added.)
            stage_exe = True
            for ft in jd.file_transfer:
                if ft.endswith(exe):
                    stage_exe = False

            if stage_exe:
                # staging from submission site to execution host is signalled by
                # using a 'site:' prefix in the staging directive'
                exe_transfer = 'site:%s > %s' % (exe, exe)
                jd.file_transfer.append(exe_transfer)

        # translate file transfer directives into actionables
        jd.transfer_directives = TransferDirectives(jd.file_transfer)


    # ----------------------------------------------------------------
    #
    def _job_run(self, jd):
        """ runs a job via condor_submit
        """

        # ensure consistency and viability of job description
        self._prepare_jd(jd)
        self._handle_file_transfers(jd.transfer_directives, mode='in')

        # create a Condor job script from SAGA job description
        script = _condorscript_generator(url=self.rm, logger=self._logger,
                jds=[jd], option_dict=self.query_options)
        self._logger.info("Generated Condor script: %s", script)

        submit_file = NamedTemporaryFile(mode='w', suffix='.condor',
                    prefix='tmp-radical-saga-', delete=False)
        submit_file_name = os.path.basename(submit_file.name)
        submit_file.write(script)
        submit_file.close()
        self._logger.info("Written Condor script locally: %s", submit_file.name)

        if self.shell.url.scheme not in ["ssh", "gsissh"]:
            raise NotImplementedError("%s support for Condor not implemented." % \
                    self.shell.url.scheme)

        self._logger.info("Transferring Condor script to: %s", self.shell.url)
        self.shell.stage_to_remote(submit_file.name, submit_file_name)

        ret, out, _ = self.shell.run_sync('%s -verbose %s' \
            % (self._commands['condor_submit'], submit_file_name))

        if ret != 0:
            # something went wrong
            raise NoSuccess("Error running job via 'condor_submit': %s."
                            "Script was: %s" % (out, script))
        else:
            # stdout contains the job id
            for line in out.split("\n"):
                if "** Proc" in line:
                    pid = line.split()[2][:-1]
                    break

            # we don't want the 'query' part of the URL to be part of the ID,
            # simply because it can get terribly long (and ugly). to get rid
            # of it, we clone the URL and set the query part to None.
            rm_clone = ru.Url (self.rm)
            rm_clone.query = ""
            rm_clone.path  = ""

            job_id = "[%s]-[%s]" % (rm_clone, pid)
            self._logger.info("Submitted Condor job with id: %s", job_id)

            # add job to internal list of known jobs.
            self.jobs[job_id]          = self._new_job_info()
            self.jobs[job_id]['state'] = PENDING
            self.jobs[job_id]['td']    = jd.transfer_directives

            # remove submit file(s)
            # TODO: leave them in case of debugging?
            self._logger.info("Submitted Condor job with scheme: '%s'", self.shell.url.scheme)
          # if self.shell.url.scheme in ['ssh', 'gsissh']:
          #     ret, out, _ = self.shell.run_sync ('rm %s' % submit_file_name)
          # else:
          #     raise NotImplementedError("%s support for Condor not implemented." % \
          #                      self.shell.url.scheme)
            os.remove(submit_file.name)

            return job_id

    # ----------------------------------------------------------------
    #
    def _retrieve_job(self, job_id):
        """ see if we can get some info about a job that we don't
            know anything about
        """
        rm, pid = self._adaptor.parse_id(job_id)

        # run the Condor 'condor_q' command to get some infos about our job
        ret, out, err = self._run_condor_q(retries=3, timeout=60, pid=pid,
                egrep=['^JobStatus', '^ExitStatus', '^CompletionDate'])

        if ret != 0:
            raise NoSuccess("Couldn't reconnect to job '%s':\n[%s]\n[%s]"
                            % (job_id, out, err))

        else:
            # the job seems to exist on the backend. let's gather some data
            job_info = self._new_job_info()

            results = out.split('\n')
            for result in results:
                if len(result.split('=')) == 2:
                    key, val = result.split('=')
                    key = key.strip()  # strip() removes whitespaces at the
                    val = val.strip()  # beginning and the end of the string

                    if   key == 'JobStatus'     : job_info['state']      = _condor_to_saga_jobstate(val)
                    elif key == 'ExitStatus'    : job_info['returncode'] = val
                    elif key == 'CompletionDate': job_info['end_time']   = val

            return job_info


    # --------------------------------------------------------------------------
    #
    def _handle_file_transfers(self, td, mode):
        """
        if mode == 'in' : perform sanity checks on all staging directives.

        if mode == 'in' : stage files to   condor submission site
        if mode == 'out': stage files from condor submission site
        """

        assert(mode in ['in', 'out'])

        # Condor data staging happens in two hops, a fact that is not reflected
        # in SAGA's file transfer directives.  The hops are:
        #
        #   1: from the application host to the condor submission site
        #   2: from the condor submission site to the execution site
        #
        # We introduce a condor specific notion to avoid the first hop on
        # staging, for those cases where we expect the files to already reside
        # on the submission site, eg. due to out-of-band staging etc.  In those
        # cases, we prefix the respective input staging source with 'site:'.
        # That prefix is also evaluated for targets of output staging, where it
        # causes the files to transfered only over the second hop, but not
        # further back to the application host.
        #
        # The code below filters for those prefixes, and will enact
        # 'stage_to_remote' and 'stage_from_remote' only for entries w/o the
        # respective prefixes.  The prefixes are removed, so that the condor
        # submission file contains clean staging directives for hop 2.

        if mode == 'in':

            td.prepared = True
            td.transfer_input_files  = []
            td.transfer_output_files = []

            # Condor specific safety checks
            if td.in_append:
                raise Exception('File append (>>) not supported')

            if td.out_append:
                raise Exception('File append (<<) not supported')

            if td.in_overwrite:

                for (local, remote) in td.in_overwrite:

                    source = local
                    target = remote

                    hop_1 = True
                    if source.startswith('site:'):
                        source = source[5:]
                        hop_1  = False

                    (s_path, s_entry) = os.path.split(source)
                    (t_path, t_entry) = os.path.split(target)

                    # make sure source is file an not dir
                    if not s_entry:
                        # TODO: this doesn't seem true, but dont want to tackle it right now
                        raise Exception('directory staging not supported: %s' % source)

                    # make sure source and target file are the same
                    if s_entry != t_entry:
                        raise Exception('source and target names must be equal: %s != %s' % (s_entry, t_entry))

                    # add for later use by job script generator
                    td.transfer_input_files.append(source)

                    if hop_1 and self.shell.url.scheme in ["ssh", "gsissh"]:
                        self._logger.info("Transferring in %s to %s", source, target)
                        self.shell.stage_to_remote(source, target,
                                                   cp_flags=sfs.CREATE_PARENTS)

            if td.out_overwrite:

                for (local, remote) in td.out_overwrite:

                    source = remote
                    target = local

                    (s_path, s_entry) = os.path.split(source)
                    (t_path, t_entry) = os.path.split(target)

                    if target.startswith('site:'):
                        target = target[5:]

                    # make sure source is file and not dir
                    if not s_entry:
                        # TODO: this doesn't seem true, but dont want to tackle it right now
                        raise Exception('directory staging not supported: %s' % source)

                    # make sure source and target file are the same
                    if s_entry != t_entry:
                        raise Exception('source and target names must be equal: %s != %s' % (s_entry, t_entry))

                    # add for later use by job script generator
                    td.transfer_output_files.append(t_entry)


        # ----------------------------------------------------------------------
        elif mode == 'out':

            # make sure mode=='in' has been used before
            if not hasattr(td, 'prepared'):
                # we don't have that attribute, its likely a cloned job
                # description or something.  Warn and bail out
                self._logger.warning('unprepared for output staging')
                return

            assert(td.prepared)

            if td.out_overwrite:

                for (local, remote) in td.out_overwrite:

                    source = remote
                    target = local

                    hop_1 = True
                    if target.startswith('site:'):
                        target = target[5:]
                        hop_1  = False

                    (s_path, s_entry) = os.path.split(source)
                    (t_path, t_entry) = os.path.split(target)

                    if hop_1 and self.shell.url.scheme in ["ssh", "gsissh"]:
                        self._logger.info("Transferring out %s to %s", source, target)
                        self.shell.stage_from_remote(source, target,
                                                     cp_flags=sfs.CREATE_PARENTS)


    # ----------------------------------------------------------------
    #
    def _job_get_info(self, job_id):
        """ get job attributes via condor_q
        """

        # if we don't have the job in our dictionary, we don't want it
        if job_id not in self.jobs:
            raise NoSuccess("Unknown job ID: %s." % job_id)

        info = self.jobs[job_id]

        # if the 'gone' flag is set, there's no need to query the job
        # state again. it's gone forever
        if info['gone'] is True:
            return info

        # if we just queried the job info, don't query again
        if time.time() - info['timestamp'] < _CACHE_TIMEOUT:
            return info

        rm, pid = self._adaptor.parse_id(job_id)

        # run the Condor 'condor_q' command to get some infos about our job
        ret, out, err = self._run_condor_q(retries=3, timeout=60, pid=pid,
                egrep=['^JobStatus', '^ExitStatus', '^CompletionDate'])

        if ret == 0:

            # parse the egrep result. this should look something like this:
            # JobStatus = 5
            # ExitStatus = 0
            # CompletionDate = 0
            results = list(filter(bool, out.split('\n')))
            for result in results:
                key, val = result.split('=', 1)
                key = key.strip()
                val = val.strip()

                if   key == 'JobStatus'     : info['state']      = _condor_to_saga_jobstate(val)
                elif key == 'ExitStatus'    : info['returncode'] = val
                elif key == 'CompletionDate': info['end_time']   = val

        else:

            # condor_q failed -- job is gone
            info['gone'] = True

            if not self._adaptor.use_hist:

                # look the other way and pray...
                info['returncode'] = 0
                info['state']      = api.DONE


            elif info['state'] in [api.RUNNING, api.PENDING]:

                # run the Condor 'condor_history' command to get info about
                # finished jobs
                self._logger.info("use condor_history on job %s", pid)
                ret, out, err = self.shell.run_sync("unset GREP_OPTIONS; %s -long -match 1 %s | \
                    grep -E '(ExitCode)|(CompletionDate)|(JobCurrentStartDate)|(QDate)|(Err)|(Out)'" \
                    % (self._commands['condor_history'], pid))

                if ret != 0:
                    # no job hist either - declar job as MIA
                    self._logger.error("Error using 'condor_history': %s : %s", out, err)
                    info['returncode'] = 1

                else:
                    # parse the egrep result. this should look something like this:
                    # ExitStatus = 0
                    results = out.split('\n')
                    for result in results:
                        if len(result.split('=')) == 2:
                            key, val = result.split('=')
                            key = key.strip()  # strip() removes whitespaces at the
                            val = val.strip()  # beginning and the end of the string

                            if   key == 'ExitCode'           : info['returncode']  = int(val)
                            elif key == 'QDate'              : info['create_time'] = val
                            elif key == 'JobCurrentStartDate': info['start_time']  = val
                            elif key == 'CompletionDate'     : info['end_time']    = val
                            elif key == 'Out'                : info['stdout']      = val.strip('"')
                            elif key == 'Err'                : info['stderr']      = val.strip('"')

                if info['returncode'] == 0: info['state'] = DONE
                else                      : info['state'] = FAILED

        if info['gone']:
            td = info['td']
            self._handle_file_transfers(td, mode='out')

        info['timestamp'] = time.time()


    # ----------------------------------------------------------------
    #
    def _job_get_info_bulk(self, cluster_id, job_ids):
        """ get job attributes via condor_q
        """

        # NOTE: bulk queries ignore the cache timeout,
        #       but they do update the timestamps

        # if the 'gone' flag is set, there's no need to query the job
        # state again. it's gone forever -- but we check all others
        found    = list()  # keep track of jobs for which we found new info
        to_check = list()  # keep track of jobs for which we found new info

        # we don't need to check final jobs
        for job_id in job_ids:

            if job_id not in self.jobs:
                raise ValueError('job %s: unknown')

            if self.jobs[job_id]['gone']:
                self._logger.debug('dont check %s', job_id)
                found.append(job_id)
            else:
                to_check.append(job_id)

        # do we have anything to do?
        if not to_check:
            return

        # run the Condor 'condor_q' command to get some infos about our job
        opts = "%s -autoformat:, ProcId JobStatus ExitStatus ExitBySignal CompletionDate" % cluster_id
        ret, out, err = self._run_condor_q(retries=3, timeout=60, options=opts)
        self._logger.debug('got state info:%s\n%s\n%s\n%s', opts, ret, out, err)

        if ret != 0:
            raise Exception("condor_q failed\n[%s]\n[%s]" % (out, err))

        results = list(filter(bool, out.split('\n')))
        ts      = time.time()
        for row in results:

            self._logger.debug('row: %s', row)

            # Some processes in cluster found with condor_q!
            procid, jobstatus, exit_code, exit_by_signal, completiondate \
                    = [col.strip() for col in row.split(',')]

            self._logger.debug('   : %s - %s - %s - %s - %s',
                    procid, jobstatus, exit_code, exit_by_signal, completiondate)

            # we always set exit_code to '1' if exited_by_signal
            if not exit_code and exit_by_signal == 'true':
                exit_code = 1

            matched = False
            for job_id in to_check:

                if job_id.endswith('.%s]' % procid):

                    self._logger.debug('match: %s', job_id)

                    found.append(job_id)
                    info = self.jobs[job_id]
                    info['state']      = _condor_to_saga_jobstate(jobstatus)
                    info['end_time']   = completiondate
                    info['returncode'] = exit_code
                    info['timestamp']  = ts
                    matched = True
                    self._logger.debug('     : %s', info['state'])
                    break

            if not matched:
                self._logger.warning('cannot match job info to any known job (%s)', row)


        # Now, see if any ids are missing, and search condor history for those:
        missing = [x for x in job_ids if x not in found]

        if self._adaptor.use_hist and missing:

            self._logger.debug('incomplete %s: %s', len(missing), missing)

            self._logger.info("use condor_history on cluster %s", cluster_id)
            cmd = "%s %s -autoformat:, " \
                  "ProcId ExitCode ExitBySignal CompletionDate " \
                  "JobCurrentStartDate QDate Err Out" \
                  % (self._commands['condor_history'], cluster_id)
            ret, out, err = self.shell.run_sync(cmd)

            if ret != 0:
                # we consider this non-fatal, as that sometimes failes on the
                # XSEDE OSG bridge without any further indication of errors
                # raise Exception("Error running 'condor_history' (%s) (%s)" % (out, err))
                self._logger.warning("condor_history failed: (%s) (%s)", out, err)
                out = ''

            else:
                results = list(filter(bool, out.split('\n')))
                ts      = time.time()
                for row in results:

                    self._logger.debug('hist row: %s', row)

                    elems = [col.strip() for col in row.split(',')]
                    if len(elems) != 8:
                        self._logger.error('condor_history noise [%s]', row)
                        continue

                    procid, exit_code, exit_by_signal,  \
                    cdate, sdate, qdate, stderr, stdout = elems

                    # we always set exit_code to '1' if exited_by_signal
                    if not exit_code and exit_by_signal == 'true':
                        exit_code = 1

                    # make sure exit code is an int:
                    try:
                        exit_code = int(exit_code)
                    except:
                        # no exit code looks wrong, we assume that condor
                        # failed, and thus also fail the job
                        self._logger.warning("condor_history w/o exit code - assume error")
                        exit_code = -1

                    matched = False
                    for job_id in to_check:

                        if job_id.endswith('.%s]' % procid):

                            matched = True

                            if job_id not in missing:
                                self._logger.debug('not missed: %s', job_id)
                                continue

                            self._logger.debug('match hist: %s: %s', job_id, elems)
                            found.append(job_id)
                            info = self.jobs[job_id]
                            info['returncode']  = int(exit_code)
                            info['create_time'] = qdate
                            info['start_time']  = sdate
                            info['end_time']    = cdate
                            info['stdout']      = stdout
                            info['stderr']      = stderr

                            if int(exit_code) == 0: info['state'] = DONE
                            else                  : info['state'] = FAILED

                            self._logger.debug('move state of %s to %s', job_id, info['state'])

                            info['gone']      = True
                            info['timestamp'] = ts
                            break

                    if not matched:
                        self._logger.warning('cannot match job info to any known job (%s)', row)


        # are still any jobs missing?
        missing = [x for x in job_ids if x not in found]
        if missing:

            # alas, condor_history seems not to work on the osg xsede bridge, so
            # we cannot consider this an error.  We will handle all remaining
            # jobs as disappeared, ie. as DONE.
            self._logger.warning('could not find all jobs %s: %s', len(missing), missing)

            ts = time.time()
            for job_id in missing:
                self._logger.warning('jobs %s disappeared', job_id)
                info = self.jobs[job_id]
                info['state']     = DONE
                info['gone']      = True
                info['timestamp'] = ts


        # FIXME: Transfer stuff.  Use self._handle_transfers(out)
        #
        # if info['gone'] is True:
        #     # If we are running over SSH, copy the output to our local system
        #     if self.shell.url.scheme in ["ssh", 'gsissh']:
        #         files = []
        #
        #         if info['transfers']:
        #             t = info['transfers']
        #             self._logger.debug("TransferOutput: %s", t)
        #
        #             # Remove leading and ending double quotes
        #             if t.startswith('"') and t.endswith('"'):
        #                 t = t[1:-1]
        #
        #             # Parse comma separated list
        #             files += t.split(',')
        #
        #         if info['stdout']:
        #             t = info['stdout']
        #             self._logger.debug("StdOut: %s", t)
        #
        #             # Remove leading and ending double quotes
        #             if t.startswith('"') and t.endswith('"'):
        #                 t = t[1:-1]
        #
        #             files.append(t)
        #
        #         if info['stderr']:
        #             t = info['stderr']
        #             self._logger.debug("StdErr: %s", t)
        #
        #             # Remove leading and ending double quotes
        #             if t.startswith('"') and t.endswith('"'):
        #                 t = t[1:-1]
        #
        #             files.append(t)
        #
        #         # Transfer list of files
        #         for f in files:
        #             f = f.strip()
        #             self._logger.info("Transferring file %s", f)
        #             self.shell.stage_from_remote(f, f)


    # ----------------------------------------------------------------
    #
    def _job_cancel_bulk(self, cluster_id, job_ids):
        """ cancel jobs via condor_rm
        """

        # NOTE: bulk queries ignore the cache timeout,
        #       but they do update the timestamps

        self._logger.debug('cancel cluster %s %s', cluster_id, job_ids)

        condor_rm = self._commands['condor_rm']

        for job_id in job_ids:

            # if we don't have the job in our dictionary, we don't want it
            if job_id not in self.jobs:
                raise ValueError("Unknown job ID: %s. Can't cancel." % job_id)

            info = self.jobs[job_id]

            if info['gone']:
                self._logger.warning("Job is already gone.")

        # run the Condor 'condor_q' command to get some infos about our job
        ret, out, _ = self.shell.run_sync("%s %s" % (condor_rm, cluster_id))

        if ret != 0:
            raise RuntimeError("condor_rm failed")

        for job_id in job_ids:
            self._logger.debug('canceled %s', job_id)
            info['state'] = CANCELED


    # ----------------------------------------------------------------
    #
    def _job_get_state(self, job_id):
        """ get the job's state
        """
        # check if we have already reach a terminal state
        if self.jobs[job_id]['state'] in [CANCELED, FAILED, DONE]:
            return self.jobs[job_id]['state']

        # check if we can / should update
        if self.jobs[job_id]['gone'] is not True:
            self._job_get_info(job_id=job_id)

        return self.jobs[job_id]['state']


    # ----------------------------------------------------------------
    #
    def _job_get_exit_code(self, job_id):
        """ get the job's exit code
        """
        # check if we can / should update
        if  (self.jobs[job_id]['gone'] is not True) and \
            (self.jobs[job_id]['returncode'] is None):
            self._job_get_info(job_id=job_id)

        ret = self.jobs[job_id]['returncode']

        # FIXME: 'None' should cause an exception
        if ret == None: return None
        else          : return int(ret)


    # ----------------------------------------------------------------
    #
    def _job_get_name(self, job_id):
        """ get the job's name
        """
        return self.jobs[job_id]['name']


    # ----------------------------------------------------------------
    #
    def _job_get_execution_hosts(self, job_id):
        """ get the job's exit code
        """
        # check if we can / should update
        if  (self.jobs[job_id]['gone']       is not True) and \
            (self.jobs[job_id]['exec_hosts'] is     None):
            self._job_get_info(job_id=job_id)

        return self.jobs[job_id]['exec_hosts']


    # ----------------------------------------------------------------
    #
    def _job_get_create_time(self, job_id):
        """ get the job's creation time
        """

        # check if we can / should update
        if  (self.jobs[job_id]['gone']        is not True) and \
            (self.jobs[job_id]['create_time'] is     None):
            self._job_get_info(job_id=job_id)

        return self.jobs[job_id]['create_time']


    # ----------------------------------------------------------------
    #
    def _job_get_start_time(self, job_id):
        """ get the job's start time
        """
        # check if we can / should update
        if  (self.jobs[job_id]['gone']       is not True) and \
            (self.jobs[job_id]['start_time'] is     None):
            self._job_get_info(job_id=job_id)

        return self.jobs[job_id]['start_time']


    # ----------------------------------------------------------------
    #
    def _job_get_end_time(self, job_id):
        """ get the job's end time
        """
        # check if we can / should update
        if  (self.jobs[job_id]['gone']     is not True) and \
            (self.jobs[job_id]['end_time'] is     None):
            self._job_get_info(job_id=job_id)

        # FIXME: convert to EPOCH
        return self.jobs[job_id]['end_time']


    # ----------------------------------------------------------------
    #
    def _job_cancel(self, job_id):
        """ cancel the job via 'condor_rm'
        """
        _, pid = self._adaptor.parse_id(job_id)

        ret, out, _ = self.shell.run_sync("%s %s\n"
            % (self._commands['condor_rm'], pid))

        if ret != 0:
            raise NoSuccess("Error canceling job via 'condor_rm': %s" % out)

        # assume the job was successfully canceled
        self.jobs[job_id]['state'] = CANCELED


    # ----------------------------------------------------------------
    #
    def _job_wait(self, job_id, timeout):
        """ wait for the job to finish or fail
        """

        time_start = time.time()
        time_now   = time_start
        rm, pid    = self._adaptor.parse_id(job_id)

        while True:
            state = self._job_get_state(job_id=job_id)

            if state in [DONE, FAILED, CANCELED]:
                    return True

            # avoid busy poll
            time.sleep(0.5)

            # check if we hit timeout
            if timeout >= 0:
                time_now = time.time()
                if time_now - time_start > timeout:
                    return False


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def create_job(self, jd):
        """ implements cpi.job.Service.get_url()
        """
        # this dict is passed on to the job adaptor class -- use it to pass any
        # state information you need there.
        adaptor_state = {"job_service":     self,
                         "job_description": jd,
                         "job_schema":      self.rm.schema,
                         "reconnect":       False
                        }

        return api.Job(_adaptor=self._adaptor, _adaptor_state=adaptor_state)



    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_job(self, jobid):
        """ Implements cpi.job.Service.get_job()
        """

        # try to get some information about this job and throw it into
        # our job dictionary.
        self.jobs[jobid] = self._retrieve_job(jobid)

        # this dict is passed on to the job adaptor class -- use it to pass any
        # state information you need there.
        adaptor_state = {"job_service":     self,
                         # TODO: fill job description
                         "job_description": api.Description(),
                         "job_schema":      self.rm.schema,
                         "reconnect":       True,
                         "reconnect_jobid": jobid
                        }

        return api.Job(_adaptor=self._adaptor, _adaptor_state=adaptor_state)



    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_url(self):
        """ implements cpi.job.Service.get_url()
        """
        return self.rm


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def list(self):
        """ implements cpi.job.Service.list()
        """

        ids = []
        ret, out, err = self._run_condor_q(retries=3, timeout=60)

        # instead of listing all jobs and grep'ing for our user ID, we only list
        # jobs submitted by us.  The result is the same, but it reduces the load
        # on the condor backend.  The drawback is that it raises and error if no
        # jobs are found - we gracefully accept that and then list no jobs,
        # which is ok, but distinguishing this from other errors is somewhat of
        # a guess.  Since we *do* see other error modes pop up from time to
        # time, we at this point ignore them all.
        if ret != 0 and len(out) > 0:
            self._logger.warning("failed to list jobs via 'condor_q': [%s] [%s]",
                              out, err)

        elif ret != 0 and len(out) == 0:
            pass

        else:

            # ------------------------------------------------------------------
            # $ /usr/bin/condor_q
            #
            #
            # -- Schedd: xd-login.opensciencegrid.org : <129.79.53.198:9615?... @ 09/27/17 07:08:42
            # OWNER  BATCH_NAME       SUBMITTED   DONE   RUN    IDLE  TOTAL JOB_IDS
            # merzky CMD: /bin/env   9/27 07:07      _      1      _      1 70722204.0
            # merzky CMD: /bin/env   9/27 07:07      _      1      _      1 70722205.0
            #
            # 2 jobs; 0 completed, 0 removed, 0 idle, 2 running, 0 held, 0 suspended
            #
            # ------------------------------------------------------------------

            for line in out.split("\n"):

                line = line.strip()

                if not line              : continue
                if line.startswith('-- '): continue
                if 'OWNER'  in line      : continue
                if 'jobs; ' in line      : continue

                elems = line.split()

                if len(elems) > 1:

                    rm_clone = ru.Url(self.rm)  # TODO: this is costly
                    rm_clone.query = ""
                    rm_clone.path  = ""

                    jobid = "[%s]-[%s]" % (rm_clone, elems[0])
                    ids.append(jobid)

        return ids


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def container_run(self, jobs):

        # the xsede condor bridge does not accept job clusters with individual
        # project IDs, thus we cluster the given jobs by project IDs, and create
        # individual submission scripts for each, and run them.
        clusters = dict()
        for job in jobs:

            jd = job.description

            # ensure consistency and viability of job description
            self._prepare_jd(jd)

            # TODO: Given that input (and output) are likely similar for
            #       bulk tasks, we probably don't want to transfer
            #       duplicates endlessly
            self._handle_file_transfers(jd.transfer_directives, mode='in')

            project = jd.project

            if not project:
                project = ''

            if project not in clusters:
                clusters[project] = list()

            clusters[project].append(job)

        for project, _jobs in clusters.items():
            self._run_cluster(project, _jobs)


    # --------------------------------------------------------------------------
    #
    def _run_cluster(self, project, jobs):

        jds = [job.description for job in jobs]

        # create a Condor job script from SAGA job description
        script = _condorscript_generator(url=self.rm, logger=self._logger,
                                         jds=jds, option_dict=self.query_options)
        self._logger.info("Generated Condor script: %s", script)

        submit_file = NamedTemporaryFile(mode='w', suffix='.condor',
                                         prefix='tmp-radica-saga-', delete=False)
        submit_file_name = os.path.basename(submit_file.name)
        submit_file.write(script)
        submit_file.close()
        self._logger.info("Written Condor script locally: %s", submit_file.name)

        if self.shell.url.scheme in ["ssh", "gsissh"]:
            self._logger.info("Transferring Condor script to: %s", self.shell.url)
            self.shell.stage_to_remote(submit_file.name, submit_file_name)

        else:
            raise NotImplementedError("%s support for Condor not implemented." %
                    self.shell.url.scheme)

        ret, out, _ = self.shell.run_sync('%s -verbose %s'
                    % (self._commands['condor_submit'], submit_file_name))

        if ret != 0:
            # condor_submit went wrong
            raise NoSuccess("Error running job via 'condor_submit': %s. "
                            "Script was: %s" % (out, script))

        pids = []
        # stdout contains the job id
        for line in out.split("\n"):
            if line.startswith("** Proc "):
                pid = line.split()[2][:-1]
                if len(pid.split('.')) != 2:
                    raise Exception("Pid looks weird: %s" % pid)
                pids.append(pid)

        if len(pids) != len(jobs):
            raise Exception("Number of pids (%d) != number of jobs (%d)"
                           % (len(pids), len(jobs)))

        # we don't want the 'query' part of the URL to be part of the ID,
        # simply because it can get terribly long (and ugly). to get rid
        # of it, we clone the URL and set the query part to None.
        rm_clone       = ru.Url(self.rm)
        rm_clone.query = ""
        rm_clone.path  = ""

        for job, pid in zip(jobs, pids):
            job_id = "[%s]-[%s]" % (rm_clone, pid)

            job._adaptor._id      = job_id
            job._adaptor._started = True

            self._logger.info("Submitted Condor job with id: %s", job_id)

            # add job to internal list of known jobs.
            self.jobs[job_id] = self._new_job_info()
            self.jobs[job_id]['state'] = PENDING
            self.jobs[job_id]['td']    = job.description.transfer_directives

        # remove submit file(s)
        # XXX: maybe leave them in case of debugging?
        # ret, out, _ = self.shell.run_sync ('rm %s' % submit_file_name)

        os.remove(submit_file.name)


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def container_wait(self, jobs, mode, timeout):

        # FIXME: implement modes
      # if mode == ANY:
      #     pass
      # if mode == ALL:
      #     pass

        while True:
            states = self.container_get_states(jobs)
            if RUNNING not in states and PENDING not in states:
                break
            time.sleep(5)

        return states[0]


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def container_cancel(self, jobs, timeout=None):

        # JobIds also include the rm, so we strip that out.  Then sort in
        # clusters we can query
        clusters = dict()
        for job in jobs:

            job_id     = job._adaptor._id
            proc_id    = self._adaptor.parse_id(job_id)[1]
            cluster_id = proc_id.split('.', 1)[0]

            if not cluster_id in clusters:
                clusters[cluster_id] = list()
            clusters[cluster_id].append(job_id)


        for cluster_id in clusters:

            job_ids   = clusters[cluster_id]
            bulk_info = self._job_cancel_bulk(cluster_id, job_ids)


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def container_get_states(self, jobs):

        log = self._logger

        # JobIds also include the rm, so we strip that out.  Then sort in
        # clusters we can query
        clusters = dict()
        for job in jobs:

            job_id     = job._adaptor._id
            proc_id    = self._adaptor.parse_id(job_id)[1]
            cluster_id = proc_id.split('.', 1)[0]

            log.debug('get bulk state for %s %s %s', job_id, proc_id, cluster_id)

            if cluster_id not in clusters:
                clusters[cluster_id] = list()
            clusters[cluster_id].append(job_id)


        states = list()
        for cluster_id in clusters:

            job_ids   = clusters[cluster_id]
            log.debug(' query job state for %s', job_ids)
            bulk_info = self._job_get_info_bulk(cluster_id, job_ids)
            states   += [self.jobs[job_id]['state'] for job_id in job_ids]

        return states

        # TODO: check "cache" for final state jobs
        # check if we have already reach a terminal state
        # if self.jobs[job_id]['state'] in [CANCELED, FAILED, DONE]:
        #     return self.jobs[job_id]['state']
        #
        # # check if we can / should update
        # if (self.jobs[job_id]['gone'] is not True):
        #     self._job_get_info(job_id=job_id)
        #
        # return self.jobs[job_id]['state']


###############################################################################
#
class CondorJob (cpi.job.Job):
    """ implements cpi.job.Job
    """

    def __init__(self, api, adaptor):

        # initialize parent class
        _cpi_base = super(CondorJob, self)
        _cpi_base.__init__(api, adaptor)


    @SYNC_CALL
    def init_instance(self, job_info):
        """ implements cpi.job.Job.init_instance()
        """
        # init_instance is called for every new api.Job object
        # that is created
        self.jd = job_info["job_description"]
        self.js = job_info["job_service"]

        self._name = self.jd.name

        # the js is responsible for job bulk operations -- which
        # for jobs only work for run()
        self._container = self.js

        if job_info['reconnect'] is True:
            self._id      = job_info['reconnect_jobid']
            self._started = True
        else:
            self._id      = None
            self._started = False

        return self.get_api()

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_description(self):
        return self.jd


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_state(self):
        """ implements cpi.job.Job.get_state()
        """
        if self._started is False:
            # jobs that are not started are always in 'NEW' state
            return NEW
        else:
            return self.js._job_get_state(self._id)


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def wait(self, timeout):
        """ implements cpi.job.Job.wait()
        """
        if self._started is False:
            raise IncorrectState("Can't wait for job that hasn't been started")

        else:
            self.js._job_wait(self._id, timeout)


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def cancel(self, timeout):
        """ implements cpi.job.Job.cancel()
        """
        if self._started is False:
            raise IncorrectState("Can't wait for job that hasn't been started")

        else:
            self.js._job_cancel(self._id)


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def run(self):
        """ implements cpi.job.Job.run()
        """
        self._id = self.js._job_run(self.jd)
        self._started = True


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_service_url(self):
        """ implements cpi.job.Job.get_service_url()
        """
        return self.js.rm


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_id(self):
        """ implements cpi.job.Job.get_id()
        """
        return self._id


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_name (self):
        """ Implements cpi.job.Job.get_name() """
        return self._name


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_exit_code(self):
        """ implements cpi.job.Job.get_exit_code()
        """
        if self._started is False:
            return None
        else:
            return self.js._job_get_exit_code(self._id)


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_created(self):
        """ implements cpi.job.Job.get_created()
        """
        if self._started is False:
            return None
        else:
            # FIXME: convert to EPOCH
            return self.js._job_get_create_time(self._id)


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_started(self):
        """ implements cpi.job.Job.get_started()
        """
        if self._started is False:
            return None
        else:
            # FIXME: convert to EPOCH
            return self.js._job_get_start_time(self._id)


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_finished(self):
        """ implements cpi.job.Job.get_finished()
        """
        if self._started is False:
            return None
        else:
            return self.js._job_get_end_time(self._id)


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_execution_hosts(self):
        """ implements cpi.job.Job.get_execution_hosts()
        """
        if self._started is False:
            return None
        else:
            return self.js._job_get_execution_hosts(self._id)


# ------------------------------------------------------------------------------





__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


""" Attribute interface """

import radical.utils            as ru
import radical.utils.signatures as rus

from . import exceptions as se

# ------------------------------------------------------------------------------

import datetime
import datetime
import traceback
import inspect
import string
import copy
import re
from   pprint import pprint

# FIXME: add a tagging 'Monitorable' interface, which enables callbacks.


now   = datetime.datetime.now
never = datetime.datetime.min

# ------------------------------------------------------------------------------
#
# define a couple of constants for the attribute API, mostly for registering
# attributes.
#
# type enums
ANY         = 'any'        # any python type can be set
URL         = 'url'        # URL type (string + URL parser checks)
INT         = 'int'        # Integer type
FLOAT       = 'float'      # float type
STRING      = 'string'     # string, duh!
BOOL        = 'bool'       # True or False or Maybe
ENUM        = 'enum'       # value is any one of a list of candidates
TIME        = 'time'       # seconds since epoch, or any py time thing
                           # which can be converted into such
                           # FIXME: conversion not implemented

# mode enums
WRITEABLE   = 'writeable'  # the consumer of the interface can change
                           # the attrib value
READONLY    = 'readonly'   # the consumer of the interface can not
                           # change the attrib value.  The
                           # implementation can still change it.
FINAL       = 'final'      # neither consumer nor implementation can
                           # change the value anymore
ALIAS       = 'alias'      # variable is deprecated, and alias'ed to
                           # a different variable.

# attrib extensions
EXTENDED    = 'extended'   # attribute added as extension
PRIVATE     = 'private'    # attribute added as private

# flavor enums
SCALAR      = 'scalar'     # the attribute value is a single data element
DICT        = 'dict'       # the attribute value is a dict of data elements
VECTOR      = 'vector'     # the attribute value is a list of data elements

# ------------------------------------------------------------------------------
#
# Callback (Abstract) Class
#
class Callback () :
    """
    Callback base class.

    All objects using the Attribute Interface allow to register a callback for
    any changes of its attributes, such as 'state' and 'state_detail'.  Those
    callbacks can be python call'ables, or derivates of this callback base
    class.  Instances which inherit this base class MUST implement (overload)
    the cb() method.

    The callable, or the callback's cb() method is what is invoked whenever the
    SAGA implementation is notified of an change on the monitored object's
    attribute.

    The cb instance receives three parameters upon invocation:


      - obj: the watched object instance
      - key:  the watched attribute (e.g. 'state' or 'state_detail')
      - val:  the new value of the watched attribute

    If the callback returns 'True', it will remain registered after invocation,
    to monitor the attribute for the next subsequent state change.  On returning
    'False' (or nothing), the callback will not be called again.

    To register a callback on a object instance, use::

      class MyCallback (saga.Callback):

          def __init__ (self):
              pass

          def cb (self, obj, key, val) :
              print(" %s\\n %s (%s) : %s"  %  self._msg, obj, key, val)

      jd  = saga.job.Description ()
      jd.executable = "/bin/date"

      js  = saga.job.Service ("fork://localhost/")
      job = js.create_job(jd)

      cb = MyCallback()
      job.add_callback(saga.STATE, cb)
      job.run()


    See documentation of the :class:`saga.Attribute` interface for further
    details and examples.
    """

    def __call__ (self, obj, key, val) :
        return self.cb (obj, key, val)

    def cb (self, obj, key, val) :
        """ This is the method that needs to be implemented by the application

            Keyword arguments::

                obj:  the watched object instance
                key:  the watched attribute
                val:  the new value of the watched attribute

            Return::

                keep:   bool, signals to keep (True) or remove (False) the callback
                        after invocation

            Callback invocation MAY (and in general will) happen in a separate
            thread -- so the application need to make sure that the callback
            code is thread-safe.

            The boolean return value is used to signal if the callback should
            continue to listen for events (return True) , or if it rather should
            get unregistered after this invocation (return False).
        """
        pass



# ------------------------------------------------------------------------------
#
class _AttributesBase (object) :
    """
    This class only exists to host properties -- as object itself does *not* have
    properties!  This class is not part of the public attribute API.
    """

    # --------------------------------------------------------------------------
    #
    @rus.takes   ('_AttributesBase')
    @rus.returns (rus.nothing)
    def __init__ (self) :
        pass


# ------------------------------------------------------------------------------
#
class Attributes (_AttributesBase, ru.DictMixin) :
    """
    Attribute Interface Class

    The Attributes interface implements the attribute semantics of the SAGA Core
    API specification (http://ogf.org/documents/GFD.90.pdf).  Additionally, this
    implementation provides that semantics the python property interface.  Note
    that a *single* set of attributes is internally managed, no matter what
    interface is used for access.

    A class which uses this interface can internally specify which attributes
    can be set, and what type they have.  Also, default values can be specified,
    and the class provides a rudimentary support for converting scalar
    attributes into vector attributes and back.

    Also, the consumer of this API can register callbacks, which get triggered
    on changes to specific attribute values.

    Example use case::


        # --------------------------------------------------------------------------------
        class Transliterator ( saga.Attributes ) :

            def __init__ (self, *args, **kwargs) :
                # setting attribs to non-extensible will cause the cal to init below to
                # complain if attributes are specified.  Default is extensible.
              # self._attributes_extensible (False)

                # pass args to base class init (requires 'extensible')
                super (Transliterator, self).__init__ (*args, **kwargs)

                # setup class attribs
                self._attributes_register   ('apple', 'Appel', URL,    SCALAR, WRITEABLE)
                self._attributes_register   ('plum',  'Pruim', STRING, SCALAR, READONLY)

                # setting attribs to non-extensible at *this* point will have allowed
                # custom user attribs on __init__ time (via args), but will then forbid
                # any additional custom attributes.
              # self._attributes_extensible (False)


        # --------------------------------------------------------------------------------
        if __name__ == "__main__":

            # define a callback method.  This callback can get registered for
            # attribute changes later.

            # ----------------------------------------------------------------------------
            def cb (key, val, obj) :
                # the callback gets information about what attribute was changed
                # on what object:
                print("called: %s - %s - %s"  %  (key, str(val), type (obj)))

                # returning True will keep the callback registered for further
                # attribute changes.
                return True
            # ----------------------------------------------------------------------------

            # create a class instance and add a 'cherry' attribute/value on
            # creation.
            trans = Transliterator (cherry='Kersche')

            # use the property interface to mess with the pre-defined
            # 'apple' attribute
            print("\\n -- apple")
            print(trans.apple)
            trans.apple = 'Abbel'
            print(trans.apple)

            # add our callback to the apple attribute, and trigger some changes.
            # Note that the callback is also triggered when the attribute's
            # value changes w/o user control, e.g. by some internal state
            # changes.
            trans.add_callback ('apple', cb)
            trans.apple = 'Apfel'

            # Setting an attribute final is actually an internal method, used by
            # the implementation to signal that no further changes on that
            # attribute are expected.  We use that here for demonstrating the
            # concept though.  Callback is invoked on set_final().
            trans._attributes_set_final ('apple')
            trans.apple = 'Abbel'
            print(trans.apple)

            # mess around with the 'plum' attribute, which was marked as
            # ReadOnly on registration time.
            print("\\n -- plum")
            print(trans.plum)
          # trans.plum    = 'Pflaume'  # raises readonly exception
          # trans['plum'] = 'Pflaume'  # raises readonly exception
            print(trans.plum)

            # check if the 'cherry' attribute exists, which got created on
            # instantiation time.
            print("\\n -- cherry")
            print(trans.cherry)

            # as we have 'extensible' set, we can add a attribute on the fly,
            # via either the property interface, or via the GFD.90 API of
            # course.
            print("\\n -- peach")
            print(trans.peach)
            trans.peach = 'Birne'
            print(trans.peach)


    This example will result in::

        -- apple
        Appel
        Appel
        Abbel
        called: apple - Abbel Appel  - <class '__main__.Transliterator'>
        called: apple - Apfel - <class '__main__.Transliterator'>
        called: apple - Apfel - <class '__main__.Transliterator'>
        Apfel

        -- plum
        Pruim
        Pruim

        -- cherry
        Kersche

        -- peach
        Berne
        Birne


    """

    # internally used constants to distinguish API from adaptor calls
    _UP    = '_up'
    _DOWN  = '_down'

    # two regexes for converting CamelCase into under_score_casing, as static
    # class vars to avoid frequent recompilation
    _camel_case_regex_1 = re.compile('(.)([A-Z][a-z]+)')
    _camel_case_regex_2 = re.compile('([a-z0-9])([A-Z])')


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  rus.anything,
                  rus.anything)
    @rus.returns (rus.nothing)
    def __init__ (self, *args, **kwargs) :
        """
        This method is not supposed to be directly called by the consumer of
        this API -- it should be called via derived object construction.

        _attributes_t_init makes sure that the basic structures are in place on
        the attribute dictionary - this saves us ton of safety checks later on.
        """
        # initialize state
        d = self._attributes_t_init ()

        # call to update and the args/kwargs handling seems to be part of the
        # dict interface conventions *shrug*
        # we use similar mechanism to initialize attribs here:

        for arg in args :
            if arg == None:
                # be resiliant to empty initialization
                pass
            elif isinstance (arg, dict):
                d['extensible']  = True   # it is just being extended ;)
                d['camelcasing'] = True   # default for dict inits
                for key in list(arg.keys()):
                    us_key = self._attributes_t_underscore(key)
                    self._attributes_i_set(us_key, arg[key], force=True, flow=self._UP)
            else:
                raise se.BadParameter("initialization expects dictionary")

        for key in list(kwargs.keys ()) :
            self.set_attribute (key, kwargs[key])

        # make iterable
        d['_iterpos'] = 0
        self.list_attributes ()



    # --------------------------------------------------------------------------
    #
    # Internal interface tools.
    #
    # These tools are only for internal use, and should never be called from
    # outside of this module.
    #
    # Naming: _attributes_t_*
    #
    @rus.takes   ('Attributes',
                  rus.optional (str))
    @rus.returns (dict)
    def _attributes_t_init (self, key=None) :
        """
        This internal function is not to be used by the consumer of this API.

        The _attributes_t_init method initializes the interface's internal data
        structures.  We always need the attribute dict, and the extensible flag.
        Everything else can be added on the fly.  The method will not overwrite
        any settings -- initialization occurs only once!

        If a key is given, the existence of this key is checked.  An exception
        is raised if the key does not exist.

        The internal data are stored as property on the _AttributesBase class.
        Storing them as property on *this* class would obviously result in
        recursion...
        """

        d = {}

        try :
            d = _AttributesBase.__getattribute__ (self, '_d')
        except :
            # need to initialize -- any exceptions in the code below should fall through
            d['attributes']  = {}
            d['extensible']  = True
            d['private']     = True
            d['camelcasing'] = False
            d['getter']      = None
            d['setter']      = None
            d['lister']      = None
            d['caller']      = None
            d['recursion']   = False
            d['_iterpos']    = 0

            _AttributesBase.__setattr__ (self, '_d', d)


        # check if we know about the given attribute
        if  key :
            if key not in d['attributes'] :
                raise se.DoesNotExist ("attribute key is invalid: %s"  %  (key))

        # all is well
        return d


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str)
    @rus.returns (str)
    def _attributes_t_keycheck (self, key) :
        """
        This internal function is not to be used by the consumer of this API.

        For the given key, check if the key name is valid, and/or if it is
        aliased.

        If the does not yet exist, the validity check is performed, and allows
        to limit dynamically added attribute names (for 'extensible' sets).

        If the key does exist, the alias check triggers a deprecation warning,
        and returns the aliased key for transparent operation.
        """

        # make sure interface is ready to use
        d = self._attributes_t_init ()

        # perform name validity checks if key is new
        if  not key in d['attributes'] :
            # FIXME: we actually don't have any tests, yet.  We should allow to
            # configure such via, say, _attributes_add_check (callable (key))
            pass


        # if key is known, check for aliasing
        else:
            # check if we know about the given attribute
            if  d['attributes'][key]['mode'] == ALIAS :
                alias = d['attributes'][key]['alias']
                print("attribute '%s' is deprecated - use '%s'"  %  (key, alias))
                key   = alias

        return key


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str,
                  rus.anything)
    @rus.returns (rus.nothing)
    def _attributes_t_call_cb (self, key, val) :
        """
        This internal function is not to be used by the consumer of this API.

        It triggers the invocation of all callbacks for a given attribute.
        Callbacks returning False (or nothing at all) will be unregistered after
        their invocation.
        """

        # make sure interface is ready to use
        d = self._attributes_t_init (key)

        # avoid recursion
        if  d['attributes'][key]['recursion'] :
            return

        callbacks = d['attributes'][key]['callbacks']

        # iterate over a copy of the callback list, so that remove does not
        # screw up the iteration
        for cb in list (callbacks) :

            call = cb

            # got the callable - call it!
            # raise and lower recursion shield as needed
            ret = False
            try :
                d['attributes'][key]['recursion'] = True
                ret = call (self, key, val)
            finally :
                d['attributes'][key]['recursion'] = False

            # remove callbacks which return 'False', or raised and exception
            if  not ret :
                callbacks.remove (cb)



    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str,
                  rus.anything)
    @rus.returns (rus.nothing)
    def _attributes_t_call_setter (self, key, val) :
        """
        This internal function is not to be used by the consumer of this API.

        It triggers the setter callbacks, to signal that the attribute value
        has just been set and should be propagated as needed.
        """

        # make sure interface is ready to use.
        d = self._attributes_t_init (key)

        # avoid recursion
        if  d['attributes'][key]['recursion'] :
            return

        # no callbacks for private keys
        if  key[0] == '_' and d['private'] :
            return

        # key_setter overwrites results from all_setter
        all_setter = d['setter']
        key_setter = d['attributes'][key]['setter']

        # Get the value via the attribute setter.  The setter will not call
        # attrib setters or callbacks, due to the recursion guard.
        # Set the value via the native setter (to the backend),
        # always raise and lower the recursion shield
        #
        # If both are present, we can ignore *one* exception.  If one
        # is present, exceptions are not ignored.
        #
        # always raise and lower the recursion shield.
        can_ignore = 0
        if  all_setter and key_setter : can_ignore = 1

        if  all_setter :
            try :
                d['attributes'][key]['recursion'] = True
                all_setter (key, val)
            except Exception as e :
                # ignoring failures from setter
                pass
            except Exception as e :
                can_ignore -= 1
                if not can_ignore : raise e
            finally :
                d['attributes'][key]['recursion'] = False

        if  key_setter :
            try :
                d['attributes'][key]['recursion'] = True
                key_setter (val)
            except:
                can_ignore -= 1
                if not can_ignore : raise
            finally :
                d['attributes'][key]['recursion'] = False



    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str)
    @rus.returns (rus.nothing)
    def _attributes_t_call_getter (self, key) :
        """
        This internal function is not to be used by the consumer of this API.

        It triggers the getter callbacks, to signal that the attribute value
        is about to be accesses and should be updated as needed.
        """

        # make sure interface is ready to use.
        d = self._attributes_t_init (key)

        # avoid recursion
        if  d['attributes'][key]['recursion'] :
            return

        # no callbacks for private keys
        if  key[0] == '_' and d['private'] :
            return

        # key getter overwrites results from all_getter
        all_getter = d['getter']
        key_getter = d['attributes'][key]['getter']


        # # Note that attributes have a time-to-live (ttl).  If a _attributes_i_get
        # # operation is attempted within 'time-of-last-update + ttl', the operation
        # # is not triggering backend getter hooks, to avoid trashing (hooks are
        # # expected to be costly).  The force flag set to True will request to call
        # # registered getter hooks even if ttl is not yet expired.
        #
        # # For example, job.wait() will update the plugin level state to 'Done',
        # # but the cached job.state attribute will remain 'New' as the plugin does
        # # not push the state change upward
        #
        # age = self._attributes_t_get_age (key)
        # ttl = d['attributes'][key]['ttl']
        #
        # if age < ttl :
        #     return


        # get the value from the native getter (from the backend), and
        # get it via the attribute getter.  The getter will not call
        # attrib setters or callbacks, due to the recursion guard.
        #
        # If both are present, we can ignore *one* exception.  If one
        # is present, exceptions are not ignored.
        #
        # always raise and lower the recursion shield.
        retries = 1
        if  all_getter and key_getter : retries = 2

        if  all_getter :
            try :
                d['attributes'][key]['recursion'] = True
                val=all_getter (key)
                d['attributes'][key]['value'] = val
            except Exception:
                retries -= 1
                if not retries : raise
            finally :
              d['attributes'][key]['recursion'] = False

        if  key_getter :
            try :
                d['attributes'][key]['recursion'] = True
                val=key_getter ()
                d['attributes'][key]['value'] = val
            except Exception:
                retries -= 1
                if not retries : raise
            finally :
                d['attributes'][key]['recursion'] = False



    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str)
    @rus.returns (rus.list_of (str))
    def _attributes_t_call_lister (self) :
        """
        This internal function is not to be used by the consumer of this API.

        It triggers the lister callback, to signal that the attribute list
        is about to be accesses and should be updated as needed.
        """

        # make sure interface is ready to use.
        d = self._attributes_t_init ()

        # avoid recursion
        if  d['recursion'] :
            return

        lister = d['lister']

        if lister :

            # the lister is simply called, and it is expected that it internally
            # adds/removes attributes as needed.
            #
            # always raise and lower the recursion shield
            try :
                d['recursion'] = True
                lister ()
            finally :
                d['recursion'] = False



    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str,
                  int,
                  callable)
    @rus.returns (rus.anything)
    def _attributes_t_call_caller (self, key, id, cb) :
        """
        This internal function is not to be used by the consumer of this API.

        It triggers the invocation of any registered caller function, usually
        after an 'add_callback()' call.
        """

        # make sure interface is ready to use.
        d = self._attributes_t_init (key)

        # avoid recursion
        if  d['recursion'] :
            return

        # no callbacks for private keys
        if  key[0] == '_' and d['private'] :
            return

        caller = d['caller']

        if  caller :

            # the caller is simply called, and it is expected that it internally
            # adds/removes callbacks as needed
            #
            # always raise and lower the recursion shield
            try :
                d['recursion'] = True
                return caller (key, id, cb)
            finally :
                d['recursion'] = False

        return



    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str)
    @rus.returns (str)
    def _attributes_t_underscore (self, key, force=False) :
        """
        This internal function is not to be used by the consumer of this API.

        The method accepts a CamelCased word, and translates that into
        'under_score' notation -- IFF 'camelcasing' is set

        Kudos: http://stackoverflow.com/questions/1175208/elegant-python-function-to-convert-camelcase-to-camel-case
        """

        # make sure interface is ready to use
        d = self._attributes_t_init ()


        if  force or d['camelcasing'] :
            temp = Attributes._camel_case_regex_1.sub(r'\1_\2', key)
            return Attributes._camel_case_regex_2.sub(r'\1_\2', temp).lower()
        else :
            return key



    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str)
    @rus.returns (rus.anything)
    def _attributes_t_conversion (self, key, val) :
        """
        This internal function is not to be used by the consumer of this API.

        The method checks a given attribute value against the attribute's
        flags, and performs some simple type conversion as needed.  Also, the
        method will restore a 'None' value to the attribute's default value.

        A deriving class can add additional value checks for attributes by
        calling :func:`_attributes_add_check` (key, check).
        """

        # make sure interface is ready to use.  We do not check for keys, that
        # needs to be done in the calling method.  For example, on 'set', type
        # conversions will be performed, but the key will not exist previously.
        d = self._attributes_t_init ()

        # if the key is not known
        if  not key in d['attributes'] :
            # cannot handle unknown attributes.  Attributes which have been
            # registered earlier will be fine, as they have type information.
            return val

        # check if a value is given.  If not, revert to the default value
        # (if available)
        if  val == None :
            if 'default' in d['attributes'][key] :
                val = d['attributes'][key]['default']


        # perform flavor and type conversion
        val = self._attributes_t_conversion_flavor (key, val)

        # apply all value checks on the conversion result
        for check in d['attributes'][key]['checks'] :
            ret = check (key, val)
            if  ret != True :
                raise se.BadParameter ("attribute value %s is not valid: %s"  %  (key, ret))

        # aaaand done
        return val


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str,
                  rus.anything)
    @rus.returns (rus.anything)
    def _attributes_t_conversion_flavor (self, key, val) :
        """
        This internal function is not to be used by the consumer of this API.
        This method should ONLY be called by _attributes_t_conversion!
        """
        # FIXME: there are possibly nicer and more reversible ways to
        #        convert the flavors...

        # easiest conversion of them all... ;-)
        if  val == None :
            return None

        # make sure interface is ready to use.
        d = self._attributes_t_init (key)

        # check if we need to serialize a list into a scalar
        f = d['attributes'][key]['flavor']
        t = d['attributes'][key]['type']
        if  f == ANY :
            # leave it alone
            return val

        elif  f == VECTOR :
            # we want a vector
            if  isinstance (val, list) :
                # val is already vec - apply type conversion on all elems
                ret = []
                for elem in val :
                    ret.append (self._attributes_t_conversion_type (key, elem))
                return ret
            else :

                # need to create vec from scalar
                if  isinstance (val, str) :

                    # for string values, we split on white spaces and type-convert
                    # all elements
                    vec = val.split ()
                    ret = []
                    for element in vec :
                        ret.append (self._attributes_t_conversion_type (key, element))
                    return ret
                else :
                    # all non-string types are interpreted as only element of
                    # a single-member list
                    return [self._attributes_t_conversion_type (key, val)]



        elif f == DICT :
            # we want a dict
            if  isinstance (val, dict) :
                # done :-)
                return val

            if  isinstance (val, list) :
                # if target type is a dict, we parse the values and
                # split on '=', creating the dict.  That will only work for
                # string typed values
                out = {}
                for elem in val :
                    (key, val) = str(elem).split ('=', 1)
                    out[key] = val
                return out

            if  isinstance (val, str) :
                # we assume a colon or comma separated list of = separated
                # key/value pairs
                elems = val.split (':')
                out   = {}
                if  len(elems) == 1 :
                    elems = val.split (',')

                for elem in elems :
                    (key, val) = str(elem).strip ().split ('=', 1)
                    out[key] = val
                return out


            # can't handle any other types...


        elif f == SCALAR :
            # we want a scalar

            if  t == ANY :
                # no need to do anything, really
                return val

            if isinstance (val, list) :
                # need to create scalar from vec
                if len (val) > 1 :
                    # if the list has more than one element, we use an intermediate
                    # string representation of the list before converting to a scalar
                    # This is the weakest conversion mode, and will not very
                    # likely yield useful results.
                    tmp = ""
                    for i in val :
                        tmp += str(i) + " "
                    return self._attributes_t_conversion_type (key, tmp)
                elif len (val) == 1 :
                    # for single element lists, we simply use the one element as
                    # scalar value
                    return self._attributes_t_conversion_type (key, val[0])
                else :
                    # no value in list
                    return None
            else :
                # scalar is already scalar, just do type conversion
                return self._attributes_t_conversion_type (key, val)


        # we should never get here...
        raise se.BadParameter ("Cannot evaluate attribute flavor (%s) : %s"  %  (key, str(f)))


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str,
                  rus.anything)
    @rus.returns (rus.anything)
    def _attributes_t_conversion_type (self, key, val) :
        """
        This internal function is not to be used by the consumer of this API.
        This method should ONLY be called by _attributes_t_conversion!
        """

        # make sure interface is ready to use.
        d = self._attributes_t_init (key)

        # oh python, how about a decent switch statement???
        t   = d['attributes'][key]['type']
        ret = None
        try :
            # FIXME: add time/date conversion to/from string
            if   t == ANY    : return        val
            elif t == INT    : return int   (val)
            elif t == FLOAT  : return float (val)
            elif t == BOOL   : return bool  (val)
            elif t == STRING : return str   (val)
            else             : return        val
        except ValueError as e:
            raise se.BadParameter ("attribute value %s has incorrect type: %s" %  (key, val)) \
                  from e

        # we should never get here...
        raise se.BadParameter ("Cannot evaluate attribute type (%s) : %s"  %  (key, str(t)))

    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str)
    @rus.returns (str)
    def _attributes_t_wildcard2regex (self, pattern) :
        """
        This internal function is not to be used by the consumer of this API.

        This method converts a string containing POSIX shell wildcards into
        a regular expression with the same matching properties::

            *       -> .*
            ?       -> .
            {a,b,c} -> (a|b|c)
            [abc]   -> [abc]
            [!abc]  -> [^abc]
        """

        re = pattern

        re.replace ('*', '.*')  # set of characters
        re.replace ('?', '.' )  # single character

        # character classes
        match = re.find ('[', 0)
        while match >= 0 :
            if  re[match + 1] == '!' :
                re[match + 1] =  '^'
            match = re.find ('[', match + 1)

        # find opening { and closing }
        first = re.find ('{', 0)
        last  = re.find ('}', first + 1)

        # while match
        while first >= 0 and last  >= 0 :
            # replace with ()
            re[first] = '('
            re[last] = '('

            # also, replace all ',' with with '|' for alternatives
            comma = re.find (',', first)
            while comma >= 0 :
                re[comma] = '|'
                comma = re.find (',', comma + 1)

            # done - find next bracket pair...
            first = re.find ('{', last + 1)
            last  = re.find ('}', first + 1)

        return re


    # --------------------------------------------------------------------------
    #
    def _attributes_t_get_age (self, key) :
        """ get the age of the attribute, i.e. seconds.microseconds since last set """

        # make sure interface is ready to use.
        d    = self._attributes_t_init (key)
        last = d['attributes'][key]['last']
        age  = now() - last

        return (age.microseconds + (age.seconds + age.days * 24 * 3600) * 1e6) / 1e6



    # --------------------------------------------------------------------------
    #
    # internal interface
    #
    # This internal interface is used by the public interfaces (dict,
    # properties, GFD.90).  We assume that CamelCasing and under_scoring is
    # sorted out before this internal interface is called.  All other tests,
    # verifications, and conversion are done here though.
    #
    # Naming: _attributes_i_*
    #
    def _attributes_i_set (self, key, val=None, force=False, flow=_DOWN) :
        """
        This internal method should not be explicitly called by consumers of
        this API, but is indirectly used via the different public interfaces.

        See :func:`set_attribute` (key, val) for details.

        New value checks can be added dynamically, and per attribute, by calling
        :func:`_attributes_add_check` (key, callable).

        Some internal methods can set the 'force' flag, and will be able to set
        attributes even in ReadOnly mode.  That is, for example, used for getter
        hooks.  Note that the Final flag will be honored even if Force is set,
        and will result in the set request being ignored.
        """

        # make sure interface is ready to use
        d = self._attributes_t_init ()

        # if the key is not known
        if not key in d['attributes'] :

            if key[0] == '_' and d['private'] :
                # if the set is private, we can register the new key.  It
                # won't have any callbacks at this point.
                self._attributes_register (key, None, ANY, ANY, WRITEABLE, EXTENDED, flow=flow)

            elif flow==self._UP or d['extensible'] :
                # if the set is extensible, we can register the new key.  It
                # won't have any callbacks at this point.
                self._attributes_register (key, None, ANY, ANY, WRITEABLE, EXTENDED, flow=flow)

            elif force :
                # someone *really* wants this attrib to be set...
                self._attributes_register (key, None, ANY, SCALAR, WRITEABLE, EXTENDED, flow=flow)

            else :
                # we cannot add new keys on non-extensible / non-private sets
                raise se.IncorrectState ("attribute set is not extensible/private (key %s)" %  key)


        # known attribute
        else :

            # check if we are allowed to change the attribute - complain if not.
            # Also, simply ignore write attempts to finalized keys.
            if 'mode' in  d['attributes'][key] :

                mode = d['attributes'][key]['mode']

                if FINAL == mode :
                    return

                elif READONLY == mode :
                    if not force :
                        raise se.BadParameter ("attribute %s is not writeable" %  key)


        # permissions are confirmed, set the attribute with conversion etc.

        # NOTE: keep the original value around for the setter
        orig_val = val

        # apply any attribute conversion
        val = self._attributes_t_conversion (key, val)

        # make sure the key's value entry exists
        if not 'value' in d['attributes'][key] :
            d['attributes'][key]['value'] = None

        # only once an attribute is explicitly set, it 'exists' for the purpose
        # of the 'attribute_exists' call, and the key iteration
        d['attributes'][key]['exists'] = True

        # # only actually change the attribute when the new value differs --
        # # and only then invoke any callbacks and hooked setters
        # if val != d['attributes'][key]['value'] :
        #
        # NOTE: this check is disabled now: we certainly want to update 'last',
        # and IMHO that should also imply a notification call, etc.  FWIW, the
        # spec is inconclusive here.
        #
        # if val != d['attributes'][key]['value'] :


        d['attributes'][key]['value'] = val
        d['attributes'][key]['last']  = now ()

        if flow==self._DOWN :
            # NOTE: we use the orig_val here, to make the environment hooks
            # happy which we introduced for BJ backward compatibility (FIXME)
            self._attributes_t_call_setter (key, orig_val)

        self._attributes_t_call_cb (key, val)


    # --------------------------------------------------------------------------
    #
    def _attributes_i_get (self, key, flow) :
        """
        This internal method should not be explicitly called by consumers of
        this API, but is indirectly used via the different public interfaces.

        see :func:`get_attribute` (key) for details.

        Note that this method is not performing any checks or conversions --
        those are all performed when *setting* an attribute.  So, any attribute
        flags (type, mode, flavor) are evaluated on setting, not on getting.
        This implementation does not account for resulting race conditions
        (changing attribute types after setting for example) -- but the public
        API does not allow that anyway.
        """

        # make sure interface is ready to use
        d = self._attributes_t_init (key)

        if flow == self._DOWN :
            self._attributes_t_call_getter (key)

        if 'value' in d['attributes'][key] :
            return d['attributes'][key]['value']

        if 'default' in d['attributes'][key] :
            return d['attributes'][key]['default']

        return None



    # --------------------------------------------------------------------------
    #
    def _attributes_i_list (self, ext=True, priv=False, CamelCase=True, flow=_DOWN) :
        """
        This internal method should not be explicitly called by consumers of
        this API, but is indirectly used via the different public interfaces.

        see :func:`list_attributes` () for details.

        Note that registration alone does not qualify for listing.  If 'ext' is
        True (default),extended attributes are listed, too.
        """

        # make sure interface is ready to use
        d = self._attributes_t_init ()

        # call list hooks to update state for listing
        self._attributes_t_call_lister ()

        ret    = []
        for key in sorted(d['attributes'].keys()) :
            if d['attributes'][key]['mode'] != ALIAS :
                if d['attributes'][key]['exists'] :

                    e = d['attributes'][key]['extended']
                    p = d['attributes'][key]['private']
                    k = key

                    if CamelCase :
                        k = d['attributes'][key]['camelcase']

                    if e and ext :
                        if p and priv :
                            ret.append (k)
                        elif not p :
                            ret.append (k)
                    elif not e :
                        if p and priv :
                            ret.append (k)
                        elif not p :
                            ret.append (k)

        return ret


    # --------------------------------------------------------------------------
    #
    def _attributes_i_find (self, pattern, flow) :
        """
        This internal method should not be explicitly called by consumers of
        this API, but is indirectly used via the different public interfaces.

        see :func:`find_attributes` (pattern) for details.
        """

        # FIXME: wildcard-to-regex

        # make sure interface is ready to use
        d = self._attributes_t_init ()


        # separate key and value pattern
        p_key  = ""    # string pattern
        p_val  = ""    # string pattern
        pc_key = None  # compiled pattern
        pc_val = None  # compiled pattern


        if pattern[0] == '=' :
            # no key pattern present, only grep on values
              p_val = self._attributes_t_wildcard2regex (pattern[1:])
        else :
          p = re.compile (r'[^\]=')
          tmp = p.split (pattern, 2)  # only split on first '='

          if len (tmp) >  0 :
              # at least one elem: only key pattern present
              p_key = self._attributes_t_wildcard2regex (tmp[0])

          if len (tmp) == 2 :
              # two elems: val pattern is also present
              p_val = self._attributes_t_wildcard2regex (tmp[1])

        # compile the found pattern
        if len (p_key) : pc_key = re.compile (p_key)
        if len (p_val) : pc_val = re.compile (p_val)

        # now dig out matching keys. List hooks are triggered in
        # _attributes_i_list(flow).
        matches = []
        for key in self._attributes_i_list (flow=flow) :
            val = str(self._attributes_i_get (key, flow=flow))

            if ( (pc_key == None) or pc_key.search (key) ) and \
               ( (pc_val == None) or pc_val.search (val) )     :
                matches.append (key)

        return matches


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str,
                  rus.one_of (_UP, _DOWN))
    @rus.returns (bool)
    def _attributes_i_exists (self, key, flow) :
        """
        This internal method should not be explicitly called by consumers of
        this API, but is indirectly used via the different public interfaces.

        see :func:`attribute_exists` (key) for details.

        Registered keys which have never been explicitly set to a value do not
        exist for the purpose of this call.
        """

        # make sure interface is ready to use
        d = self._attributes_t_init ()

        # check if we know about that attribute
        if key in d['attributes'] :
            if 'exists' in d['attributes'][key] :
                if  d['attributes'][key]['exists'] :
                    return True

        return False


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str,
                  rus.one_of (_UP, _DOWN))
    @rus.returns (bool)
    def _attributes_i_is_extended (self, key, flow) :
        """
        This internal method should not be explicitly called by consumers of
        this API, but is indirectly used via the different public interfaces.

        This method will check if the given key is extended, i.e. was registered
        on the fly, vs. registered explicitly.

        This method is not used by, and not exposed via the public API, yet.
        """

        # make sure interface is ready to use
        d = self._attributes_t_init (key)

        return d['attributes'][key]['extended']


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str,
                  rus.one_of (_UP, _DOWN))
    @rus.returns (bool)
    def _attributes_i_is_private (self, key, flow) :
        """
        This internal method should not be explicitly called by consumers of
        this API, but is indirectly used via the different public interfaces.

        This method will check if the given key is private, i.e. starts with an
        underscore and 'allow_private' is enabled.

        This method is not used by, and not exposed via the public API, yet.
        """

        # make sure interface is ready to use
        d = self._attributes_t_init (key)

        return d['attributes'][key]['private']


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str,
                  rus.one_of (_UP, _DOWN))
    @rus.returns (bool)
    def _attributes_i_is_readonly (self, key, flow) :
        """
        This internal method should not be explicitly called by consumers of
        this API, but is indirectly used via the different public interfaces.

        see L{attribute_is_readonly} (key) for details.

        This method will check if the given key is readonly, i.e. cannot be
        'set'.  The call will also return 'True' if the attribute is final
        """

        # make sure interface is ready to use
        d = self._attributes_t_init (key)

        # check if we know about that attribute
        if  d['attributes'][key]['mode'] == FINAL or \
            d['attributes'][key]['mode'] == READONLY :
            return True

        return False


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str,
                  rus.one_of (_UP, _DOWN))
    @rus.returns (bool)
    def _attributes_i_is_writeable (self, key, flow) :
        """
        This internal method should not be explicitly called by consumers of
        this API, but is indirectly used via the different public interfaces.

        see :func:`attribute_is_writable` (key) for details.

        This method will check if the given key is writeable - i.e. not readonly.
        """

        return not self._attributes_i_is_readonly (key, flow=flow)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str,
                  rus.one_of (_UP, _DOWN))
    @rus.returns (bool)
    def _attributes_i_is_removable (self, key, flow) :
        """
        This internal method should not be explicitly called by consumers of
        this API, but is indirectly used via the different public interfaces.

        see :func:`attribute_is_removable` (key) for details.

        'True' if the attrib is writeable and Extended.
        """

        if self._attributes_i_is_writeable (key, flow=flow) and \
           self._attributes_i_is_extended  (key, flow=flow)     :
            return True

        return False

    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str,
                  rus.one_of (_UP, _DOWN))
    @rus.returns (bool)
    def _attributes_i_is_vector (self, key, flow) :
        """
        This internal method should not be explicitly called by consumers of
        this API, but is indirectly used via the different public interfaces.

        see :func:`attribute_is_vector` (key) for details.
        """

        # make sure interface is ready to use
        d = self._attributes_t_init (key)

        # check if we know about that attribute
        if  d['attributes'][key]['flavor'] == VECTOR :
            return True

        return False


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str,
                  rus.one_of (_UP, _DOWN))
    @rus.returns (bool)
    def _attributes_i_is_final (self, key, flow) :
        """
        This internal method should not be explicitly called by consumers of
        this API, but is indirectly used via the different public interfaces.

        This method will query the 'final' flag for an attribute, which signals
        that the attribute will never change again.

        This method is not used by, and not exposed via the public API, yet.
        """

        # make sure interface is ready to use
        d = self._attributes_t_init (key)

        if FINAL == d['attributes'][key]['mode'] :
             return True

        # no final flag found -- assume non-finality!
        return False


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str,
                  callable,
                  rus.one_of (_UP, _DOWN))
    @rus.returns (int)
    def _attributes_i_add_cb (self, key, cb, flow) :
        """
        This internal method should not be explicitly called by consumers of
        this API, but is indirectly used via the different public interfaces.

        see :func:`add_callback` (key, cb) for details.
        """

        # make sure interface is ready to use
        d = self._attributes_t_init (key)

        d['attributes'][key]['callbacks'].append (cb)

        id = len (d['attributes'][key]['callbacks']) - 1

        if flow==self._DOWN :
            self._attributes_t_call_caller (key, id, cb)

        return id


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str,
                  rus.optional (int),
                  rus.one_of (_UP, _DOWN))
    @rus.returns (rus.nothing)
    def _attributes_i_del_cb (self, key, id=None, flow=_DOWN) :
        """
        This internal method should not be explicitly called by consumers of
        this API, but is indirectly used via the different public interfaces.

        see :func:`remove_callback` (key, cb) for details.
        """

        # make sure interface is ready to use
        d = self._attributes_t_init (key)

        if flow==self._DOWN :
            self._attributes_t_call_caller (key, id, None)

        # id == None: remove all callbacks
        if not id :
            d['attributes'][key]['callbacks'] = []
        else :
            if len (d['attributes'][key]['callbacks']) < id :
                raise se.BadParameter ("invalid callback cookie for attribute %s"  %  key)
            else :
                # do not pop from list, that would invalidate the id's!
                d['attributes'][key]['callbacks'][id] = None



    # --------------------------------------------------------------------------
    #
    # This part of the interface is primarily for use in deriving classes, which
    # thus provide the Attributes interface.
    #
    # Keys should be provided as CamelCase (only relevant if camelcasing is
    # set).
    #
    # Naming: _attributes_*
    #
    @rus.takes   ('Attributes',
                  str,
                  rus.optional (rus.optional (rus.anything)),
                  rus.optional (rus.one_of (ANY, URL, INT, FLOAT, STRING, BOOL, ENUM, TIME)),
                  rus.optional (rus.one_of (ANY, SCALAR, VECTOR, DICT)),
                  rus.optional (rus.one_of (READONLY, WRITEABLE, ALIAS, FINAL)),
                  rus.optional (rus.one_of (bool, EXTENDED)),
                  rus.optional (rus.one_of (_UP, _DOWN)))
    @rus.returns (rus.nothing)
    def _attributes_register (self, key,      default=None, typ=ANY, flavor=ANY,
                              mode=WRITEABLE, ext=False,    flow=_DOWN) :
        """
        This interface method is not part of the public consumer API, but can
        safely be called from within derived classes.

        Register a new attribute.

        This function ignores extensible, private, final and readonly flags.  It
        can also be used to re-register an existing attribute with new
        properties -- the old attribute value, callbacks etc. will be lost
        though.  Using this call that way may result in confusing behaviour on
        the public API level.
        """
        # FIXME: check for valid mode and flavor settings

        # make sure interface is ready to use
        d = self._attributes_t_init ()

        priv = False
        if d['private'] and key[0] == '_' :
            priv = True

        # we expect keys to be registered as CamelCase (in those cases where
        # that matters).  But we store attributes in 'under_score' version.
        us_key = self._attributes_t_underscore (key)

        # retain old values
        val    = default
        exists = False

        if  default != None :
            exists = True

        if us_key in d['attributes'] :
            val    = d['attributes'][us_key]['value']
            exists = True

        # register the attribute and properties
        d['attributes'][us_key]                 = {}
        d['attributes'][us_key]['value']        = val     # initial value
        d['attributes'][us_key]['default']      = default # default value
        d['attributes'][us_key]['type']         = typ     # int, float, enum, ...
        d['attributes'][us_key]['exists']       = exists  # no value set, yet?
        d['attributes'][us_key]['flavor']       = flavor  # scalar / vector
        d['attributes'][us_key]['mode']         = mode    # readonly / writeable / final
        d['attributes'][us_key]['extended']     = ext     # is an extended attribute
        d['attributes'][us_key]['private']      = priv    # is a  private attribute
        d['attributes'][us_key]['camelcase']    = key     # keep original key name
        d['attributes'][us_key]['underscore']   = us_key  # keep under_scored name
        d['attributes'][us_key]['enums']        = []      # list of valid enum values
        d['attributes'][us_key]['checks']       = []      # list of custom value checks
        d['attributes'][us_key]['callbacks']    = []      # list of callbacks
        d['attributes'][us_key]['recursion']    = False   # recursion check for callbacks
        d['attributes'][us_key]['setter']       = None    # custom attribute setter
        d['attributes'][us_key]['getter']       = None    # custom attribute getter
        d['attributes'][us_key]['last']         = never   # time of last refresh (never)
        d['attributes'][us_key]['ttl']          = 0.0     # refresh delay (none)

        # for enum types, we add a value checker
        if typ == ENUM :

            def _enum_check (key, val) :
                if None == val  :
                    # None is always allowed
                    return True

                us_key = self._attributes_t_underscore (key)
                d      = self._attributes_t_init       (us_key)

                vals   = d['attributes'][us_key]['enums']

                # check if there is anything to check
                if not vals :
                    return True

                # value must be one of allowed enums
                if val in vals :
                        return True

                # Houston, we got a problem...
                msg = "incorrect value (%s) for Enum typed attribute (%s)." \
                      "Allowed values: %s"  %  (str(val), key, str(vals))
                raise se.BadParameter (msg)

            self._attributes_add_check (key, _enum_check, flow=flow)



    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str,
                  str,
                  rus.one_of (_UP, _DOWN))
    @rus.returns (rus.nothing)
    def _attributes_register_deprecated (self, key, alias, flow=_DOWN) :
        """
        Often enough, there is the need to use change attribute names.  It is
        good practice to not simply rename attributes, and thus effectively
        remove old ones, as that is likely to break existing code.  Instead, new
        names are added, and old names are kept for a certain time for backward
        compatibility.  To support migration to the new names, the old names
        should be marked as 'deprecated' though - which can be configured to
        print a warning whenever an old, deprecated attribute is used.

        This method allows to register such deprecated attribute names.  They
        can thus be used just like new ones, and in fact are implemented as
        aliases to the new ones -- but they will print a deprecated warning on
        usage.

        The first parameter is the old name of the attribute, the second
        parameter is the aliased new name.  Note that the new name needs to be
        registered before (via :class:`saga._attributes_register`)::

            # old code:
            self._attributes_register ('apple', 'Appel', STRING, SCALAR, WRITEABLE)

            # new code
            self._attributes_register ('fruit', 'Appel', STRING, SCALAR, WRITEABLE)
            self._attributes_register_deprecated ('apple', 'fruit')

        In some cases, you may want to deprecate a variable and not replace it
        with a new one.  In order to keep this interface simple, this can be
        achieved via::

            # new code
            self._attributes_register ('deprecated_apple', 'Appel', STRING, SCALAR, WRITEABLE)
            self._attributes_register_deprecated ('apple', 'deprecated_apple')

        This way, the user will either see a warning, or has to explicitly use
        'deprecated_apple' as attribute name -- which should be warning enough,
        at least for the programmer ;o)
        """

        # we expect keys to be registered as CamelCase (in those cases where
        # that matters).  But we store attributes in 'under_score' version.
        us_alias = self._attributes_t_underscore (alias)
        us_key   = self._attributes_t_underscore (key)

        # make sure interface is ready to use
        # This check will throw if 'alias' was not registered before.
        d = self._attributes_t_init (us_alias)

        # remove any old instance of this attribute
        if us_key in  d['attributes'] :
            self._attributes_unregister (us_key, flow=flow)

        # register the attribute and properties
        d['attributes'][us_key]               = {}
        d['attributes'][us_key]['mode']       = ALIAS      # alias
        d['attributes'][us_key]['alias']      = us_alias   # aliased var
        d['attributes'][us_key]['camelcase']  = key        # keep original key name
        d['attributes'][us_key]['underscore'] = us_key     # keep under_scored name



    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str,
                  rus.one_of (_UP, _DOWN))
    @rus.returns (rus.nothing)
    def _attributes_unregister (self, key, flow) :
        """
        This interface method is not part of the public consumer API, but can
        safely be called from within derived classes.

        Unregister an attribute.

        This function ignores the extensible, private, final and readonly flag,
        and is supposed to be used by derived classes, not by the consumer of
        the API.

        Note that unregistering is different from setting the value to 'None' --
        all meta information about the attribute will be removed.  Further
        attempts to access the attribute from the public API will result in an
        DoesNotExist exception.  This method should be used sparingly -- in
        fact, GFD.90 requires final attributes to stay around forever (frozen).
        """

        # make sure interface is ready to use
        us_key = self._attributes_t_underscore (key)
        d      = self._attributes_t_init       (us_key)

        # if the attribute exists, purge it
        if us_key in d['attributes'] :
            del (d['attributes'][us_key])


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str,
                  rus.one_of (_UP, _DOWN))
    @rus.returns (rus.nothing)
    def _attributes_remove (self, key, flow) :
        """
        This interface method is not part of the public consumer API, but can
        safely be called from within derived classes.

        Remove an extended an attribute.

        This function allows to safely remove any attribute which is 'private'
        or 'extended' and has write permissions.
        """

        # make sure interface is ready to use
        us_key = self._attributes_t_underscore (key)
        d      = self._attributes_t_init       (us_key)

        if self._attributes_i_is_removable (key, flow=flow) :
            del (d['attributes'][us_key])


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str,
                  rus.optional (rus.list_of (rus.anything)),
                  rus.optional (rus.one_of  (_UP, _DOWN)))
    @rus.returns (rus.nothing)
    def _attributes_set_enums (self, key, enums=None, flow=_DOWN) :
        """
        This interface method is not part of the public consumer API, but can
        safely be called from within derived classes.

        Specifies the set of allowed values for Enum typed attributes.  If not
        set, or if list is None, any values are allowed.
        """

        us_key = self._attributes_t_underscore (key)
        d      = self._attributes_t_init       (us_key)

        d['attributes'][us_key]['enums'] = enums


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  rus.optional (bool),
                  rus.optional (callable),
                  rus.optional (callable),
                  rus.optional (callable),
                  rus.optional (callable),
                  rus.optional (rus.one_of  (_UP, _DOWN)))
    @rus.returns (rus.nothing)
    def _attributes_extensible (self, e=True,
                                getter=None, setter=None,
                                lister=None, caller=None,
                                flow=_DOWN) :
        """
        This interface method is not part of the public consumer API, but can
        safely be called from within derived classes.

        Allow (or forbid) the on-the-fly creation of new attributes.  Note that
        this method also allows to *remove* the extensible flag -- that leaves
        any previously created extended attributes untouched, but just prevents
        the creation of new extended attributes.
        """

        d = self._attributes_t_init ()
        d['extensible'] = e

        if getter : self._attributes_set_global_getter (getter, flow=flow)
        if setter : self._attributes_set_global_setter (setter, flow=flow)
        if lister : self._attributes_set_global_lister (lister, flow=flow)
        if caller : self._attributes_set_global_caller (caller, flow=flow)



    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  rus.optional (bool),
                  rus.optional (rus.one_of  (_UP, _DOWN)))
    @rus.returns (rus.nothing)
    def _attributes_allow_private (self, p=True, flow=_DOWN) :
        """
        This interface method is not part of the public consumer API, but can
        safely be called from within derived classes.

        Allow (or forbid) the on-the-fly creation of private attributes
        (starting with underscore).  Note that this method also allows to
        *remove* the respective flag -- that leaves any previously created
        private attributes untouched, but just prevents the creation of new
        private attributes.
        """

        d = self._attributes_t_init ()
        d['private'] = p


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  rus.optional (bool),
                  rus.optional (rus.one_of  (_UP, _DOWN)))
    @rus.returns (rus.nothing)
    def _attributes_camelcasing (self, c=True, flow=_DOWN) :
        """
        This interface method is not part of the public consumer API, but can
        safely be called from within derived classes.

        Use 'CamelCase' for dict entries and the GFD.90 API, but 'under_score'
        for properties.

        Note that we do not provide an option to turn CamelCasing off - once it
        is turned on, it stays on -- otherwise we would loose attributes...
        """

        d = self._attributes_t_init ()
        d['camelcasing'] = c


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  'Attributes',
                  rus.optional (rus.one_of  (_UP, _DOWN)))
    @rus.returns ('Attributes')
    def _attributes_deep_copy (self, other, flow=_DOWN) :
        """
        This interface method is not part of the public consumer API, but can
        safely be called from within derived classes.

        This method can be used to make sure that deep copies of derived classes
        are also deep copies of the respective attributes.  In accordance with
        GFD.90, the deep copy will ignore callbacks.  It will copy checks
        though, as the assumption is that value constraints stay valid.

        Note that we don't copy private keys.
        """


        # make sure interface is ready to use
        d = self._attributes_t_init ()

        other_d = {}
        orig_d  = other._attributes_t_init ()

        # for some reason, deep copy won't work on the 'attributes' dict, so we
        # do it manually.  Use the list copy c'tor to copy list elements.
        other_d['camelcasing']  = d['camelcasing']
        other_d['attributes']   = d['attributes']
        other_d['extensible']   = d['extensible']
        other_d['private']      = d['private']
        other_d['camelcasing']  = d['camelcasing']
        other_d['recursion']    = d['recursion']
        other_d['getter']       = d['setter']
        other_d['setter']       = d['setter']
        other_d['lister']       = d['lister']
        other_d['caller']       = d['caller']

        other_d['attributes'] = {}

        for key in d['attributes'] :
            other_d['attributes'][key] = {}
            other_d['attributes'][key]['default']      =       d['attributes'][key]['default']
            other_d['attributes'][key]['exists']       =       d['attributes'][key]['exists']
            other_d['attributes'][key]['type']         =       d['attributes'][key]['type']
            other_d['attributes'][key]['flavor']       =       d['attributes'][key]['flavor']
            other_d['attributes'][key]['mode']         =       d['attributes'][key]['mode']
            other_d['attributes'][key]['extended']     =       d['attributes'][key]['extended']
            other_d['attributes'][key]['private']      =       d['attributes'][key]['private']
            other_d['attributes'][key]['camelcase']    =       d['attributes'][key]['camelcase']
            other_d['attributes'][key]['underscore']   =       d['attributes'][key]['underscore']
            other_d['attributes'][key]['enums']        = list (d['attributes'][key]['enums'])
            other_d['attributes'][key]['checks']       = list (d['attributes'][key]['checks'])
            other_d['attributes'][key]['callbacks']    = list (d['attributes'][key]['callbacks'])
            other_d['attributes'][key]['recursion']    =       d['attributes'][key]['recursion']
            other_d['attributes'][key]['setter']       =       d['attributes'][key]['setter']
            other_d['attributes'][key]['getter']       =       d['attributes'][key]['getter']
            other_d['attributes'][key]['last']         =       d['attributes'][key]['last']
            other_d['attributes'][key]['ttl']          =       d['attributes'][key]['ttl']

            if d['attributes'][key]['private' ] and key in orig_d['attributes'] :
                # don't copy private keys
                other_d['attributes'][key] = orig_d['attributes'][key]

            else :
                other_d['attributes'][key]['value']  = copy.deepcopy (d['attributes'][key]['value'])

        # set the new dictionary as state for copied class
        _AttributesBase.__setattr__ (other, '_d', other_d)

        return other


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  ('Attributes', dict))
    @rus.returns ('Attributes')
    def __deepcopy__ (self, memo) :
        other = Attributes ()
        return self._attributes_deep_copy (other)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  rus.optional (str),
                  rus.optional (rus.one_of  (_UP, _DOWN)))
    @rus.returns (rus.nothing)
    def _attributes_dump (self, msg=None, flow=_DOWN) :
        """
        This interface method is not part of the public consumer API, but can
        safely be called from within derived classes.

        Debugging dump to stdout.
        """

        # make sure interface is ready to use
        d = self._attributes_t_init ()


        keys_all = sorted (d['attributes'].keys ())

        print("---------------------------------------")
        print(str (type (self)))

        if msg :
            print("---------------------------------------")
            print(msg)

        print("---------------------------------------")
        print(" %-30s : %s"  %  ("Extensible"  , d['extensible']))
        print(" %-30s : %s"  %  ("Private"     , d['private']))
        print(" %-30s : %s"  %  ("CamelCasing" , d['camelcasing']))
        print("---------------------------------------")

        keys_exist = []
        for key in keys_all :
            if  'exists' in d['attributes'][key] and \
                d['attributes'][key]['exists']   :
                keys_exist.append (key)

        print("'Registered' attributes")
        for key in keys_all :
            if key not in keys_exist :
                if not  d['attributes'][key]['mode'] == ALIAS and \
                   not  d['attributes'][key]['extended'] :
                    print(" %-30s [%6s, %6s, %9s, %3d]: %s"  % \
                             (d['attributes'][key]['camelcase'],
                              d['attributes'][key]['type'],
                              d['attributes'][key]['flavor'],
                              d['attributes'][key]['mode'],
                          len(d['attributes'][key]['callbacks']),
                              d['attributes'][key]['value']
                              ))

        print("---------------------------------------")

        print("'Existing' attributes")
        keys_exist.sort ()
        for key in keys_exist :
            if not  d['attributes'][key]['mode'] == ALIAS :
                print(" %-30s [%6s, %6s, %9s, %3d]: %s"  % \
                         (d['attributes'][key]['camelcase'],
                          d['attributes'][key]['type'],
                          d['attributes'][key]['flavor'],
                          d['attributes'][key]['mode'],
                      len(d['attributes'][key]['callbacks']),
                          d['attributes'][key]['value']
                          ))

        print("---------------------------------------")

        print("'Extended' attributes")
        for key in keys_all :
            if key not in keys_exist :
                if not  d['attributes'][key]['mode'] == ALIAS and \
                        d['attributes'][key]['extended'] :
                    print(" %-30s [%6s, %6s, %9s, %3d]: %s"  % \
                             (d['attributes'][key]['camelcase'],
                              d['attributes'][key]['type'],
                              d['attributes'][key]['flavor'],
                              d['attributes'][key]['mode'],
                          len(d['attributes'][key]['callbacks']),
                              d['attributes'][key]['value']
                              ))

        print("---------------------------------------")

        print("'Deprecated' attributes (aliases)")
        for key in keys_all :
            if key not in keys_exist :
                if d['attributes'][key]['mode'] == ALIAS :
                    print(" %-30s [%24s]:  %s"  % \
                             (d['attributes'][key]['camelcase'],
                              ' ',
                              d['attributes'][key]['alias']
                              ))

        print("---------------------------------------")


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str,
                  rus.optional (rus.anything),
                  rus.optional (rus.one_of (_UP, _DOWN)))
    @rus.returns (rus.nothing)
    def _attributes_set_final (self, key, val=None, flow=_DOWN) :
        """
        This interface method is not part of the public consumer API, but can
        safely be called from within derived classes.

        This method will set the 'final' flag for an attribute, signalling that
        the attribute will never change again.  The ReadOnly flag is ignored.
        A final value can optionally be provided -- otherwise the attribute is
        frozen with its current value.

        Note that attributes_set_final() will trigger callbacks, even if the
        value was not set, or did not change.
        """

        # make sure interface is ready to use
        us_key = self._attributes_t_underscore (key)
        d      = self._attributes_t_init       (us_key)

        newval = val
        oldval = d['attributes'][us_key]['value']
        if None == newval :
            # freeze at current value unless indicated otherwise
            val = oldval

        # flag as final, and set the final value (this order to avoid races in
        # callbacks)
        d['attributes'][us_key]['mode'] = FINAL
        self._attributes_i_set (us_key, val, flow=flow)

        # callbacks are not invoked if the value did not change -- we take care
        # of that here.
        #
        # if  None == newval or oldval == newval :

        self._attributes_t_call_cb (key, val)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str,
                  rus.optional (float),
                  rus.optional (rus.one_of (_UP, _DOWN)))
    @rus.returns (rus.nothing)
    def _attributes_set_ttl (self, key, ttl=0.0, flow=_DOWN) :
        """ set attributes TTL in seconds (float) -- see L{_attributes_i_set} """

        # make sure interface is ready to use.
        us_key = self._attributes_t_underscore (key)
        d      = self._attributes_t_init       (us_key)

        d['attributes'][us_key]['ttl'] = ttl



    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str,
                  callable,
                  rus.optional (rus.one_of (_UP, _DOWN)))
    @rus.returns (rus.nothing)
    def _attributes_add_check (self, key, check, flow=_DOWN) :
        """
        This interface method is not part of the public consumer API, but can
        safely be called from within derived classes.

        Value checks can be added dynamically, and per attribute.  'callable'
        needs to be a python callable, and will be invoked as::

            callable (key, val)

        Those checks will be invoked whenever a new attribute value is set.  If
        that call then returns 'True', the value is accepted.  Otherwise, the
        value will be considered to be invalid, which results in an exception as
        per above.  'callable' can return a string as error message.
        """

        # make sure interface is ready to use
        us_key = self._attributes_t_underscore (key)
        d = self._attributes_t_init (us_key)

        # register the attribute and properties
        d['attributes'][us_key]['checks'].append (check)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str,
                  callable,
                  rus.optional (rus.one_of (_UP, _DOWN)))
    @rus.returns (rus.nothing)
    def _attributes_set_getter (self, key, getter, flow=_DOWN) :
        """
        This interface method is not part of the public consumer API, but can
        safely be called from within derived classes.

        The Attribute API makes no assumptions on how attribute values are kept
        up-to-date.  In general, it expects an asynchronous thread to set
        attribute values as needed.  To keep the complexity low for backend
        developers, it also supports the registration of 'setter' and 'getter'
        hooks.  Those are very similar to callbacks, but kind of inversed: where
        frontend callbacks are invoked on backend attribute changes, backend
        hooks are invoked on frontend attribute queries.  They are expected to
        internally trigger state updates.

        For example, on::

            print(c.attrib)

        The attribute getter for the 'attrib' attribute will be invoked.  If for
        that attribute a getter hook is registered, that hook will first query
        the backend for value updates.  After that update has been performed,
        the getter will retrieve the (updated) value.

        Similarly, setter hooks will be invoked *after* the attribute setter
        method, to inform the implementation of the updated attribute value.

        Further, list hooks are invoked before a list or find operation is
        really internally executed, to allow the implementation to updated the
        list of available attributes.

        Note that only one setter/getter/lister method can be registered (for
        setters/getters per key, for listers per class instance).

        Hooks have a different call signature than callbacks::

            setter        (self, value)
            getter        (self)
            global_setter (self, key, value)
            global_setter (self, key)
            global_lister (self)
            global_caller (self, key)

        Note that frequent setter, and even more list and getter calls are very
        common -- the implementation of hooks should consider to cache the
        respective values.
        """

        # make sure interface is ready to use
        us_key = self._attributes_t_underscore (key)
        d      = self._attributes_t_init       (us_key)

        # register the attribute and properties
        d['attributes'][us_key]['getter'] = getter


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str,
                  callable,
                  rus.optional (rus.one_of (_UP, _DOWN)))
    @rus.returns (rus.nothing)
    def _attributes_set_setter (self, key, setter, flow=_DOWN) :
        """
        This interface method is not part of the public consumer API, but can
        safely be called from within derived classes.

        See documentation of L{_attributes_set_getter } for details.
        """

        # make sure interface is ready to use
        us_key = self._attributes_t_underscore (key)
        d      = self._attributes_t_init       (us_key)

        # register the attribute and properties
        d['attributes'][us_key]['setter'] = setter


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  callable,
                  rus.optional (rus.one_of (_UP, _DOWN)))
    @rus.returns (rus.nothing)
    def _attributes_set_global_lister (self, lister, flow) :
        """
        This interface method is not part of the public consumer API, but can
        safely be called from within derived classes.

        See documentation of L{_attributes_set_getter } for details.
        """

        d = self._attributes_t_init ()

        # register the attribute and properties
        d['lister'] = lister


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  callable,
                  rus.optional (rus.one_of (_UP, _DOWN)))
    @rus.returns (rus.nothing)
    def _attributes_set_global_caller (self, caller, flow) :
        """
        This interface method is not part of the public consumer API, but can
        safely be called from within derived classes.

        See documentation of :class:`saga._attributes_set_setter ` for details.
        """

        d = self._attributes_t_init ()

        # register the attribute and properties
        d['caller'] = caller


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  callable,
                  rus.optional (rus.one_of (_UP, _DOWN)))
    @rus.returns (rus.nothing)
    def _attributes_set_global_getter (self, getter, flow=_DOWN) :
        """
        This interface method is not part of the public consumer API, but can
        safely be called from within derived classes.

        See documentation of L{_attributes_set_getter } for details.
        """

        d = self._attributes_t_init ()

        # register the attribute and properties
        d['getter'] = getter


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  callable,
                  rus.optional (rus.one_of (_UP, _DOWN)))
    @rus.returns (rus.nothing)
    def _attributes_set_global_setter (self, setter, flow) :
        """
        This interface method is not part of the public consumer API, but can
        safely be called from within derived classes.

        See documentation of L{_attributes_set_getter } for details.
        """

        d = self._attributes_t_init ()

        # register the attribute and properties
        d['setter'] = setter


    # --------------------------------------------------------------------------
    #
    # the GFD.90 attribute interface
    #
    # The GFD.90 interface supports CamelCasing, and thus converts all keys to
    # underscore before using them.
    @rus.takes   ('Attributes',
                  str,
                  rus.anything,
                  rus.optional (rus.one_of (_UP, _DOWN)))
    @rus.returns (rus.nothing)
    def set_attribute (self, key, val, _flow=_DOWN) :
        """
        set_attribute(key, val)

        This method sets the value of the specified attribute.  If that
        attribute does not exist, DoesNotExist is raised -- unless the attribute
        set is marked 'extensible' or 'private'.  In that case, the attribute is
        created and set on the fly (defaulting to mode=writeable, flavor=Scalar,
        type=ANY, default=None).  A value of 'None' may reset the attribute to
        its default value, if such one exists (see documentation).

        Note that this method is performing a number of checks and conversions,
        to match the value type to the attribute properties (type, mode, flavor).
        Those conversions are not guaranteed to yield the expected result -- for
        example, the conversion from 'scalar' to 'vector' is, for complex types,
        ambiguous at best, and somewhat stupid.  The consumer of the API SHOULD
        ensure correct attribute values.  The conversions are intended to
        support the most trivial and simple use cases (int to string etc).
        Failed conversions will result in an BadParameter exception.

        Attempts to set a 'final' attribute are silently ignored.  Attempts to
        set a 'readonly' attribute will result in an IncorrectState exception
        being raised.

        Note that set_attribute() will trigger callbacks, if a new value
        (different from the old value) is given.
        """

        key    = self._attributes_t_keycheck   (key)
        us_key = self._attributes_t_underscore (key)
        return   self._attributes_i_set        (us_key, val, flow=_flow)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str,
                  rus.optional (rus.one_of (_UP, _DOWN)))
    @rus.returns (rus.anything)
    def get_attribute (self, key, _flow=_DOWN) :
        """
        get_attribute(key)

        This method returns the value of the specified attribute.  If that
        attribute does not exist, an DoesNotExist is raised.  It is not an
        error to query an existing, but unset attribute though -- that will
        result in 'None' to be returned (or the default value, if available).
        """

        key    = self._attributes_t_keycheck   (key)
        us_key = self._attributes_t_underscore (key)
        return   self._attributes_i_get        (us_key, _flow)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str,
                  rus.list_of (rus.anything),
                  rus.optional (rus.one_of (_UP, _DOWN)))
    @rus.returns (rus.nothing)
    def set_vector_attribute (self, key, val, _flow=_DOWN) :
        """
        set_vector_attribute (key, val)

        See also: :func:`saga.Attributes.set_attribute` (key, val).

        As python can handle scalar and vector types transparently, this method
        is in fact not very useful.  For that reason, it maps internally to the
        set_attribute method.
        """

        key    = self._attributes_t_keycheck   (key)
        us_key = self._attributes_t_underscore (key)
        return   self._attributes_i_set        (us_key, val, _flow)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str,
                  rus.optional (rus.one_of (_UP, _DOWN)))
    @rus.returns (rus.list_of (rus.anything))
    def get_vector_attribute (self, key, _flow=_DOWN) :
        """
        get_vector_attribute (key)

        See also: :func:`saga.Attributes.get_attribute` (key).

        As python can handle scalar and vector types transparently, this method
        is in fact not very useful.  For that reason, it maps internally to the
        get_attribute method.
        """

        key    = self._attributes_t_keycheck   (key)
        us_key = self._attributes_t_underscore (key)
        return   self._attributes_i_get        (us_key, _flow)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str,
                  rus.optional (rus.one_of (_UP, _DOWN)))
    @rus.returns (rus.nothing)
    def remove_attribute (self, key, _flow=_DOWN) :
        """
        remove_attribute (key)

        Removing an attribute is actually different from unsetting it, or from
        setting it to 'None'.  On remove, all traces of the attribute are
        purged, and the key will not be listed on
        :func:`saga.Attributes.list_attributes` () anymore.
        """

        key    = self._attributes_t_keycheck   (key)
        us_key = self._attributes_t_underscore (key)
        return   self._attributes_remove       (us_key, _flow)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  rus.optional (rus.one_of (_UP, _DOWN)))
    @rus.returns (rus.list_of (str))
    def list_attributes (self, _flow=_DOWN) :
        """
        list_attributes ()

        List all attributes which have been explicitly set.
        """

        return self._attributes_i_list (flow=_flow)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str,
                  rus.optional (rus.one_of (_UP, _DOWN)))
    @rus.returns (rus.list_of (str))
    def find_attributes (self, pattern, _flow=_DOWN) :
        """
        find_attributes (pattern)

        Similar to list(), but also grep for a given attribute pattern.  That
        pattern is of the form 'key=val', where both 'key' and 'val' can contain
        POSIX shell wildcards.  For non-string typed attributes, the pattern is
        applied to a string serialization of the typed value, if that exists.
        """

        return self._attributes_i_find (pattern, _flow)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str,
                  rus.optional (rus.one_of (_UP, _DOWN)))
    @rus.returns (bool)
    def attribute_exists (self, key, _flow=_DOWN) :
        """
        attribute_exist (key)

        This method will check if the given key is known and was set explicitly.
        The call will also return 'True' if the value for that key is 'None'.
        """

        key    = self._attributes_t_keycheck   (key)
        us_key = self._attributes_t_underscore (key)
        return self._attributes_i_exists (us_key, _flow)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str,
                  rus.optional (rus.one_of (_UP, _DOWN)))
    @rus.returns (bool)
    def attribute_is_readonly (self, key, _flow=_DOWN) :
        """
        attribute_is_readonly (key)

        This method will check if the given key is readonly, i.e. cannot be
        'set'.  The call will also return 'True' if the attribute is final
        """

        key    = self._attributes_t_keycheck   (key)
        us_key = self._attributes_t_underscore (key)
        return self._attributes_i_is_readonly (us_key, _flow)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str,
                  rus.optional (rus.one_of (_UP, _DOWN)))
    @rus.returns (bool)
    def attribute_is_writeable (self, key, _flow=_DOWN) :
        """
        attribute_is_writeable (key)

        This method will check if the given key is writeable - i.e. not readonly.
        """

        key    = self._attributes_t_keycheck   (key)
        us_key = self._attributes_t_underscore (key)
        return self._attributes_i_is_writeable (us_key, _flow)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str,
                  rus.optional (rus.one_of (_UP, _DOWN)))
    @rus.returns (bool)
    def attribute_is_removable (self, key, _flow=_DOWN) :
        """
        attribute_is_writeable (key)

        This method will check if the given key can be removed.
        """

        key    = self._attributes_t_keycheck   (key)
        us_key = self._attributes_t_underscore (key)
        return self._attributes_i_is_removable (us_key, _flow)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str,
                  rus.optional (rus.one_of (_UP, _DOWN)))
    @rus.returns (bool)
    def attribute_is_vector (self, key, _flow=_DOWN) :
        """
        attribute_is_vector (key)

        This method will check if the given attribute has a vector value type.
        """

        key    = self._attributes_t_keycheck   (key)
        us_key = self._attributes_t_underscore (key)
        return self._attributes_i_is_vector (us_key, _flow)


    # --------------------------------------------------------------------------
    #
    # fold the GFD.90 monitoring API into the attributes API
    #
    @rus.takes   ('Attributes',
                  str,
                  callable,
                  rus.optional (rus.one_of (_UP, _DOWN)))
    @rus.returns (int)
    def add_callback (self, key, cb, _flow=_DOWN) :
        """
        add_callback (key, cb)

        For any attribute change, the API will check if any callbacks are
        registered for that attribute.  If so, those callbacks will be called
        in order of registration.  This registration function will return an
        id (cookie) identifying the callback -- that id can be used to
        remove the callback.

        A callback is any callable python construct, and MUST accept three
        arguments::

            - STRING key: the name of the attribute which changed
            - ANY    val: the new value of the attribute
            - ANY    obj: the object on which this attribute interface was called

        The 'obj' can be any python object type, but is guaranteed to expose
        this attribute interface.

        The callback SHOULD return 'True' or 'False' -- on 'True', the callback
        will remain registered, and will thus be called again on the next
        attribute change.  On returning 'False', the callback will be
        unregistered, and will thus not be called again.  Returning nothing is
        interpreted as 'False', other return values lead to undefined behavior.

        Note that callbacks will not be called on 'Final' attributes (they will
        be called once as that attribute enters finality).
        """

        key    = self._attributes_t_keycheck   (key)
        us_key = self._attributes_t_underscore (key)
        return self._attributes_i_add_cb (us_key, cb, _flow)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str,
                  int,
                  rus.optional (rus.one_of (_UP, _DOWN)))
    @rus.returns (rus.nothing)
    def remove_callback (self, key, id, _flow=_DOWN) :
        """
        remove_callback (key, id)

        This method allows to unregister a previously registered callback, by
        providing its id.  It is not an error to remove a non-existing cb, but
        a valid ID MUST be provided -- otherwise, a BadParameter is raised.

        If no ID is provided (id == None), all callbacks are removed for this
        attribute.
        """

        key    = self._attributes_t_keycheck   (key)
        us_key = self._attributes_t_underscore (key)
        return self._attributes_i_del_cb (us_key, id, _flow)



    # --------------------------------------------------------------------------
    #
    # Python property interface
    #
    # we assume that properties are always used in under_score notation.
    #
    @rus.takes   ('Attributes',
                  str)
    @rus.returns (rus.anything)
    def __getattr__ (self, key) :
        """ see L{get_attribute} (key) for details. """

        key  = self._attributes_t_keycheck (key)
        return self._attributes_i_get      (key, flow=self._DOWN)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str,
                  rus.anything)
    @rus.returns (rus.nothing)
    def __setattr__ (self, key, val) :
        """ see L{set_attribute} (key, val) for details. """

        key  = self._attributes_t_keycheck (key)
        return self._attributes_i_set      (key, val, flow=self._DOWN)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  str)
    @rus.returns (rus.nothing)
    def __delattr__ (self, key) :
        """ see L{remove_attribute} (key) for details. """

        key  = self._attributes_t_keycheck (key)
        return self._attributes_remove     (key, flow=self._DOWN)

    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes')
    @rus.returns (str)
    def __str__  (self) :
        """ return a string representation of all set attributes """

        s = "%s %s" % (type(self), str(self.as_dict()))

        return s


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes',
                  dict)
    @rus.returns (dict)
    def from_dict (self, seed):
        """ set attributes from dict """

        for k,v in seed.items():
            self.set_attribute(k,v)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Attributes')
    @rus.returns (dict)
    def as_dict (self) :
        """ return a dict representation of all set attributes """

        d = {}

        for a in self.list_attributes () :
            d[a] = self.get_attribute (a)

        return d


    # --------------------------------------------------------------------------
    #
    # Python dictionary interface, via the DictMixin
    #
    # we assume that keys are always used in under_score notation.
    #
    # --------------------------------------------------------------------------
    #
    def __getitem__ (self, key) :
        return self.get_attribute(key)

    # --------------------------------------------------------------------------
    #
    def __setitem__ (self, key, value) :
        return self.set_attribute (key, value)

    # --------------------------------------------------------------------------
    #
    def __delitem__ (self, key) :
        return self.remove_attribute (key)

    # --------------------------------------------------------------------------
    #
    def keys (self) :
        return self._attributes_i_list (CamelCase=False)

    # --------------------------------------------------------------------------
    #
    def __iter__ (self) :
        return self

    # --------------------------------------------------------------------------
    #
    def __next__ (self) :

        iterlist = self._attributes_i_list (CamelCase=False)

        d = self._attributes_t_init ()

        if  d['_iterpos'] >= len(iterlist) :
            d['_iterpos']  = 0
            raise StopIteration

        if  not len(iterlist) :
            d['_iterpos']  = 0
            raise StopIteration

        d['_iterpos'] += 1

        return iterlist[d['_iterpos']-1]



# ------------------------------------------------------------------------------

# FIXME: add
#   - class metric()
#   - add_metric()
#   - remove_metric()
#   - fire_metric()
#   - list_metrics()
#   - get_metric()
#   - list_callbacks()

# ------------------------------------------------------------------------------







__author__    = "Andre Merzky"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


""" Provides the SAGA Job Service CPI """

from ..decorators import CPI_SYNC_CALL  as SYNC
from ..decorators import CPI_ASYNC_CALL as ASYNC
from ..base       import CPIBase
from ..sasync     import Async


class Service (CPIBase, Async) :

    # ----------------------------------------------------------------
    #
    # initialization methods
    #
    def __init__ (self, api, adaptor) :

        _cpi_base = super  (Service, self)
        _cpi_base.__init__ (api, adaptor)

    @SYNC
    def init_instance              (self, rm, session)         : pass
    @ASYNC
    def init_instance_async        (self, rm, session)         : pass

    @SYNC
    def close                      (self)                      : pass


    # ----------------------------------------------------------------
    #
    # job service methods
    #
    @SYNC
    def create_job                 (self, jd, ttype)           : pass
    @ASYNC
    def create_job_async           (self, jd, ttype)           : pass

    @SYNC
    def run_job                    (self, cmd, host, ttype)    : pass
    @ASYNC
    def run_job_async              (self, cmd, host, ttype)    : pass

    @SYNC
    def list                       (self, ttype)               : pass
    @ASYNC
    def list_async                 (self, ttype)               : pass

    @SYNC
    def get_url                    (self, ttype)               : pass
    @ASYNC
    def get_url_async              (self, ttype)               : pass

    @SYNC
    def get_job                    (self, job_id, ttype)       : pass
    @ASYNC
    def get_job_async              (self, job_id, ttype)       : pass

    @SYNC
    def get_self                   (self, ttype)               : pass
    @ASYNC
    def get_self_async             (self, ttype)               : pass

    @SYNC
    def container_run              (self, jobs)                : pass
    @ASYNC
    def container_run_async        (self, jobs)                : pass

    @SYNC
    def container_wait             (self, jobs, mode, timeout) : pass
    @ASYNC
    def container_wait_async       (self, jobs, mode, timeout) : pass

    @SYNC
    def container_cancel           (self, jobs, timeout)       : pass
    @ASYNC
    def container_cancel_async     (self, jobs, timeout)       : pass

    @SYNC
    def container_get_states       (self, jobs)                : pass
    @ASYNC
    def container_get_states_async (self, jobs)                : pass




